from typing import (
    Sequence, Set, AbstractSet, List, Optional, Type, Union, Iterable,
    Mapping, Any)
import logging
import collections

from neuron_morphology.feature_extractor.mark import Mark
from neuron_morphology.feature_extractor.marked_feature import (
    MarkedFeature, Feature
)
from neuron_morphology.feature_extractor.feature_extraction_run import \
    FeatureExtractionRun
from neuron_morphology.feature_extractor.data import Data


# The register_features method on FeatureExtractor supports one level of
# nesting in its inputs
RegistrableFeature = Union[
    Feature, Mapping[Any, Feature], Iterable[Feature]
]

class FeatureExtractor:

    def __init__(self, features: Sequence[Feature] = tuple()):
        """ Extracts morphological features from data

        Parameters
        ----------
        features : a sequence of marked (ideally) callables, each of which
            defines some feature calculation. These define the set of features
            eligible to be calculated on any specific run

        """

        self.marks: Set[Type[Mark]] = set()
        self.features: List[MarkedFeature] = []

        if features:
            self.register_features(features)

    def register_features(self, features: Sequence[RegistrableFeature]):
        """ Add a new feature to the list of options

        Parameters
        ----------
        features : the features to be registered. If it is not already marked,
            it will be registered with no marks

        """
        for feature in features:
            to_register: List[Feature] = []

            if isinstance(feature, collections.abc.Mapping):
                to_register.extend(feature.values())
            elif isinstance(feature, collections.abc.Iterable):
                to_register.extend(feature)
            else:
                to_register.append(feature)

            for feature_to_register in to_register:
                feature_to_register = MarkedFeature.ensure(feature_to_register)
                self.marks |= feature_to_register.marks
                self.features.append(feature_to_register)

        return self

    def extract(
        self,
        data: Data,
        only_marks: Optional[AbstractSet[Type[Mark]]] = None,
        required_marks: AbstractSet[Type[Mark]] = frozenset()
    ) -> FeatureExtractionRun:
        """ Run the feature extractor for a single dataset

        Parameters
        ----------
        data : the dataset from which features will be calculated
        only_marks : if provided, reject marks not in this set
        required_marks : if provided, raise an exception if any of these marks
            do not validate successfully

        Returns
        -------
        The calculated features, along with a record of the marks and features
            selected.

        """

        return (
            FeatureExtractionRun(data)
                .select_marks(
                    self.marks,
                    required_marks=required_marks
                )
                .select_features(
                    self.features,
                    only_marks=only_marks
                )
                .extract()
        )



import unittest

import math
from neuron_morphology.features import soma
from neuron_morphology.constants import (
    SOMA, AXON, APICAL_DENDRITE)
from neuron_morphology.morphology import Morphology
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.feature_extractor.feature_extractor import (
    FeatureExtractor
)
from neuron_morphology.feature_extractor.marked_feature import (
    specialize
)
from neuron_morphology.feature_extractor.feature_specialization import (
    BasalDendriteSpec, AxonSpec
)



def basic_nodes():
    """
    This morphology looks like:
    S -10-> A -10-> A
    |
    3-> AD -3-> AD
    """
    return [
        {
            "id": 0,
            "parent_id": -1,
            "type": SOMA,
            "x": 0,
            "y": 0,
            "z": 100,
            "radius": 5
        },
        {
            "id": 1,
            "parent_id": -1,
            "type": AXON,
            "x": 1,
            "y": 5,
            "z": 110,
            "radius": 1
        },
        {
            "id": 2,
            "parent_id": 1,
            "type": AXON,
            "x": 3,
            "y": 10,
            "z": 120,
            "radius": 1
        },
        {
            "id": 3,
            "parent_id": 0,
            "type": APICAL_DENDRITE,
            "x": 5,
            "y": 15,
            "z": 100,
            "radius": 1
        },
        {
            "id": 4,
            "parent_id": 3,
            "type": APICAL_DENDRITE,
            "x": 7,
            "y": 20,
            "z": 100,
            "radius": 1
        },
    ]


class MorphSomaTest(unittest.TestCase):
    def setUp(self):
        self.morphology = Morphology(
            basic_nodes(),
            node_id_cb=lambda node: node["id"],
            parent_id_cb=lambda node: node["parent_id"],
        )
        self.data = Data(self.morphology, relative_soma_depth= 0.25)

class TestSomaFeatures(MorphSomaTest):
    
    def test_soma_surface(self):
        obtained = soma.calculate_soma_surface(self.data)
        self.assertAlmostEqual(obtained, 4.0 * math.pi * 5 * 5)

    def test_relative_soma_depth(self):
        obtained = soma.calculate_relative_soma_depth(self.data)
        self.assertAlmostEqual(obtained, 0.25)

    def test_stem_exit_and_distance(self):
        first, second = soma.calculate_stem_exit_and_distance(self.data, [AXON])[0]
        self.assertAlmostEqual(first, 0.447, 3)
        self.assertAlmostEqual(second, 11.22, 2)

    def test_stem_exit_with_type(self):

        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0, radius=10)
                    .basal_dendrite(10, 0, 0, radius=2)
                        .basal_dendrite(20, 0, 0, radius=2).up()
                        .axon(30, 0, 0, radius=0.5)
                            .axon(40, 0, 0, radius=0.5).up(3)
                    .basal_dendrite(0, -10, 0, radius=2)
                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(soma.calculate_stem_exit_and_distance, [BasalDendriteSpec, AxonSpec])])
        feature_extraction_run = fe.extract(cell_data)

        axon_results = feature_extraction_run.results["axon.calculate_stem_exit_and_distance"]
        basal_results = feature_extraction_run.results["basal_dendrite.calculate_stem_exit_and_distance"]
        self.assertEqual(axon_results[0][0], 0.5)
        self.assertEqual(axon_results[0][1], 30.0)
        self.assertEqual(basal_results[0][0], 0.5)
        self.assertEqual(basal_results[0][1], 0)
        self.assertEqual(basal_results[1][0], 1)
        self.assertEqual(basal_results[1][1], 0)

    def test_number_of_stems_with_type(self):

        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .axon().up()
                    .basal_dendrite().up()
                    .basal_dendrite()
                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(soma.calculate_number_of_stems, [BasalDendriteSpec])])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(
            feature_extraction_run.results["basal_dendrite.calculate_number_of_stems"],
            2
        )





class TestSomaPercentile(unittest.TestCase):
    def setUp(self):
        # morphology with 3/4 axons below soma in z
        self.morphology = Morphology([
            {
                "id": 0,
                "parent_id": -1,
                "type": SOMA,
                "x": 0,
                "y": 0,
                "z": 100,
                "radius": 5
            },
            {
                "id": 1,
                "parent_id": 0,
                "type": AXON,
                "x": 0,
                "y": 0,
                "z": 110,
                "radius": 1
            },
            {
                "id": 2,
                "parent_id": 1,
                "type": AXON,
                "x": 0,
                "y": 0,
                "z": 90,
                "radius": 1
            },
            {
                "id": 3,
                "parent_id": 2,
                "type": AXON,
                "x": 0,
                "y": 0,
                "z": 80,
                "radius": 1
            },
            {
                "id": 4,
                "parent_id": 3,
                "type": AXON,
                "x": 0,
                "y": 0,
                "z": 70,
                "radius": 1
            },
        ],
            node_id_cb=lambda node: node["id"],
            parent_id_cb=lambda node: node["parent_id"],
        )

        self.data = Data(self.morphology)

    def test_soma_percentile(self):
        obtained = soma.soma_percentile(self.data, [AXON])
        self.assertEqual(obtained[2], 0.25)

    def test_soma_percentile_no_sym(self):
        obtained = soma.soma_percentile(self.data, [AXON], symmetrize_xz=False)
        self.assertEqual(obtained[2], 0.75)



from unittest import TestCase

import xarray as xr
import numpy as np
from shapely.geometry.polygon import Polygon
from shapely.geometry import LineString
from scipy.interpolate import RegularGridInterpolator

import neuron_morphology.layered_point_depths.__main__ as lpd

class TestUtilities(TestCase):

    def test_translate_field(self):
        inp = xr.DataArray(
            data=np.eye(10),
            dims=["x", "y"],
            coords = {
                "x": np.arange(10),
                "y": np.arange(10, 20)
            }
        )

        res = lpd.translate_field(inp, 5, 2)

        assert np.allclose(res.coords["x"], np.arange(5, 15))
        assert np.allclose(res.coords["y"], np.arange(12, 22))

    def test_setup_interpolator(self):

        field = xr.DataArray(
            data=np.arange(16).reshape((4, 4)),
            dims=["x", "y"],
            coords = {
                "x": np.arange(4),
                "y": np.arange(4, 8)
            }
        )

        obt = lpd.setup_interpolator(field, dim=None, method="linear")
        self.assertAlmostEqual(
            obt((2.5, 6.5)),
            12.5
        ) 

    def test_containing_layer(self):
        layers = [
            {
                "name": "a",
                "bounds": Polygon([(0, 0), (1, 0), (1, 1), (0, 1), (0, 0)])
            },
            {
                "name": "b",
                "bounds":  Polygon([(1, 1), (2, 1), (2, 2), (1, 2), (1, 1)])
            }
        ]

        self.assertEqual(
            lpd.containing_layer((1.5, 1.5), layers),
            "b"
        )

    def test_tuplize(self):
        arr = np.array([1, 2, 3])
        self.assertEqual(
            lpd.tuplize(arr),
            (1, 2, 3)
        )

    def test_step_from_node(self):
        pos = (0, 0)

        depth_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.arange(16).reshape((4, 4))
        )

        dx_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.ones((4, 4))
        )
        dy_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.ones((4, 4)) * 2
        )

        surface = LineString([(0, 2), (4, 2)])

        depth = lpd.step_from_node(
            pos, depth_interp, dx_interp, dy_interp, surface, 1.0, 1000)

        self.assertAlmostEqual(
            depth, 
            depth_interp((1, 2))
        )

    def test_get_node_intersections(self):

        node = {
            "x": 0,
            "y": 0,
            "z": 12,
            "id": 100,
            "type": "foo"
        }

        depth_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.arange(16).reshape((4, 4))
        )

        dx_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.ones((4, 4))
        )
        dy_interp = RegularGridInterpolator(
            (np.arange(4), np.arange(4)),
            np.ones((4, 4)) * 2
        )

        layers = [
            {
                "name": "a",
                "bounds": Polygon([(0, 0), (0, 2), (4, 2), (4, 0), (0, 0)]),
                "pia_surface": LineString([(0, 2), (4, 2)]),
                "wm_surface": LineString([(0, 0), (4, 0)])
            }
        ]

        obt = lpd.get_node_intersections(
            node, 
            depth_interp, 
            dx_interp, 
            dy_interp, 
            layers, 
            1.0, 
            1000
        )
        

        self.assertEqual(obt["ids"], node["id"])
        self.assertEqual(obt["point_type"], node["type"])
        self.assertEqual(obt["layer_name"], "a")
        self.assertAlmostEqual(
            obt["local_layer_pia_side_depth"], 
            depth_interp((1, 2))
        )
        self.assertAlmostEqual(
            obt["local_layer_wm_side_depth"], 
            depth_interp((0, 0))
        )

    def test_setup_layers(self):
        bounds = [(0, 0), (0, 2), (4, 2), (4, 0), (0, 0)]
        pia = [[0, 2], [4, 2]]
        wm = [(0, 0), (4, 0)]
        layers = [
            {
                "name": "a",
                "bounds": Polygon(bounds),
                "pia_surface": LineString(pia),
                "wm_surface": LineString(wm)
            }
        ]

        lpd.setup_layers(layers)
        assert np.allclose(
            list(layers[0]["bounds"].exterior.coords),
            bounds
        )


import unittest

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.marked_feature import (
    specialize
)
from neuron_morphology.feature_extractor.feature_specialization import (
    NEURITE_SPECIALIZATIONS
)
from neuron_morphology.features.path import (
    max_path_distance, mean_contraction
)


class TestPath(unittest.TestCase):

    def test_contraction(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 1, 0).up()
                                .basal_dendrite(3, 1, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(2, -2, 0)
                                    .axon(1, -2, 0).up()
                                    .axon(3, -2, 0).up(5)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 2, 0).up()
                            .apical_dendrite(1, 2, 0)
                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(mean_contraction, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["axon.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["dendrite.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.mean_contraction"], 1)

    def test_max_path_distance(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .axon(1, -1, 0).up()
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(3, 0, 0).up()
                            .basal_dendrite(2, 1, 0).up(3)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 2, 0).up()
                            .apical_dendrite(1, 2, 0)
                                .apical_dendrite(1, 3, 0)
                .build()
        )


        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(max_path_distance, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.max_path_distance"], 4)
        self.assertEqual(feature_extraction_run.results["axon.max_path_distance"], 2)
        self.assertEqual(feature_extraction_run.results["dendrite.max_path_distance"], 4)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.max_path_distance"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.max_path_distance"], 4)



from typing import Dict, Optional, List, Any, IO, Union

import logging
import copy as cp

import numpy as np
from scipy.spatial.distance import euclidean
import pandas as pd
import h5py

from argschema.argschema_parser import ArgSchemaParser
from neuron_morphology.transforms.tilt_correction._schemas import (
    InputParameters, OutputParameters)
from neuron_morphology.morphology import Morphology
from neuron_morphology.swc_io import morphology_from_swc
import neuron_morphology.transforms.affine_transform as aff


CCF_SHAPE = (1320, 800, 1140)
CCF_RESOLUTION = 10


def get_tilt_correction(morphology: Morphology,
                        soma_voxel: List[int],
                        slice_angle_matrix: float,
                        closest_path,
                        ):
    """
        Find the tilt angle between the slice plane and the nearest streamline

        Parameters
        ----------
        morphology: Morphology object
        soma_voxel: soma voxel in ccf {'x': , 'y': , 'z': }
        slice_angle_matrix: 4 x 4 affine matrix of the slice plane
            relative to ccf
        closest_path: 3 x N array of voxel coordinates in closest streamline,
                      only first (wm end) and last (pia end) coordinates are used.
                      In future the tilt correction may be refined
                      to use the entire path.

        Returns
        -------
        tilt angle correction (radians)

    """

    # Find slice plane vector
    M = slice_angle_matrix[0:3, :]
    o_ccf = np.dot(M, np.array([0, 0, 0, 1]))  # slice plane origin in ccf
    x_ccf = np.dot(M, np.array([1, 0, 0, 1]))  # slice plane x vector in ccf
    y_ccf = np.dot(M, np.array([0, 1, 0, 1]))  # slice plane y vector in ccf

    norm_vec = np.cross(x_ccf - o_ccf, y_ccf - o_ccf)  # z vector normal to slice plane
    norm_unit = norm_vec / euclidean(norm_vec, [0, 0, 0])

    # Find approximate streamline vector
    if euclidean(closest_path[:, 0], soma_voxel) == 0:
        # if soma is at the top, calculate from other end of streamline
        streamline_unit = (soma_voxel - closest_path[:, -1]) /\
            euclidean(soma_voxel, closest_path[:, -1])
    else:
        streamline_unit = (closest_path[:, 0] - soma_voxel) /\
             euclidean(closest_path[:, 0], soma_voxel)

    # determine angle between norm and streamline
    # using dot(a,b) = norm(a)*norm(b)*cos(theta)
    # and norm(cross(a,b)) = norm(a)*norm(b)*sin(theta)
    # and tan(theta) = sin(theta)/cos(theta)
    # therefore tan(theta) = norm(cross(a,b)) / dot(a,b)
    norm_cross = np.linalg.norm(np.cross(norm_unit, streamline_unit))
    dot_prod = np.dot(norm_unit, streamline_unit)
    theta = np.arctan2(norm_cross, dot_prod)

    tilt_angle = np.pi / 2 - theta

    return tilt_angle


def find_closest_path(soma_voxel: List[int],
                      ccf_path: Union[str, IO],
                      n_sublists: int = 2):
    '''
        Finds closest path to soma_voxel

        Parameters
        ----------
        soma_voxel: List containing soma voxel ccf coordinates
        ccf_path: str, FilePath, or File-like object openable by H5PY
        n_sublists: will seperate path_ids into n sublists to load in
                    at a time. Higher values decrease memory usage but
                    increase processing time. n = 1 uses about 16GB

        Returns
        -------
        closest_path: array of voxel coordinates of the closest streamline
    '''
    soma_voxel = np.reshape(soma_voxel, (3, 1))

    closest_path_id = None
    min_distance = np.inf
    with h5py.File(ccf_path, 'r') as ccf_data:
        path_ids = np.arange(ccf_data['paths'].shape[0])
        sublists = np.array_split(path_ids, n_sublists)
        for sublist in sublists:
            paths = ccf_data['paths'][sublist, :]
            for sublist_idx, path_id in enumerate(sublist):
                path = paths[sublist_idx, :]
                path = path[path > 0]
                if path is None:
                    continue

                voxels = np.array(np.unravel_index(path, CCF_SHAPE))
                delta = voxels - soma_voxel
                distances = np.linalg.norm(delta, axis=0)
                distance = distances.min()

                if distance < min_distance:
                    min_distance = distance
                    closest_path_id = path_id
                    if distance < 1:
                        return voxels

        closest_path = ccf_data['paths'][closest_path_id, :]

    closest_path = closest_path[closest_path > 0]
    closest_path_coords = np.array(np.unravel_index(closest_path, CCF_SHAPE))

    return closest_path_coords


def determine_slice_flip(morphology: Morphology,
                         soma_marker: Dict,
                         slice_image_flip: bool):
    """
        Determines whether the tilt correction should be positive or negative

        Parameters
        ----------
        morphology: Morphology object
        soma_marker: soma marker dictionary from reconstruction marker file
        slice_image_flip: indicates whether the image was flipped relative
                          to the slice (e.g the z axis of the image is opposite
                          to the z axis in the slice)

        Returns
        -------
        flip_toggle -1 or 1 to be multiplied against tilt correction
    """

    flip_toggle = 1

    morph_soma = morphology.get_soma()
    if (soma_marker['z'] - morph_soma['z']) > 0:
        flip_toggle = -1

    if flip_toggle == 1 and not slice_image_flip:
        flip_toggle *= -1

    return flip_toggle


def read_soma_marker(marker_path: str):
    col_names = ['x', 'y', 'z', 'radius', 'shape', 'name',
                 'comment', 'color_r', 'color_g', 'color_b']
    markers = pd.read_csv(marker_path, sep=',', comment='#',
                          header=None, names=col_names)
    soma_marker = markers.loc[markers['name'] == 30].to_dict('r')
    return soma_marker[0]


def run_tilt_correction(
        morphology: Morphology,
        soma_marker: Dict,
        ccf_soma_location: Dict,
        slice_transform: aff.AffineTransform,
        slice_image_flip: bool,
        ccf_path: Union[str, IO]):

    soma_voxel = (int(ccf_soma_location["x"] // CCF_RESOLUTION),
                  int(ccf_soma_location["y"] // CCF_RESOLUTION),
                  int(ccf_soma_location["z"] // CCF_RESOLUTION))

    closest_path = find_closest_path(soma_voxel,
                                     ccf_path)

    tilt = get_tilt_correction(morphology,
                               soma_voxel,
                               slice_transform.affine,
                               closest_path)

    flip_toggle = determine_slice_flip(morphology,
                                       soma_marker,
                                       slice_image_flip)

    tilt = tilt * flip_toggle

    transform = aff.AffineTransform(
                    aff.affine_from_transform(
                        aff.rotation_from_angle(tilt, axis=0)))

    return tilt, transform


def main():
    parser = ArgSchemaParser(
        schema_type=InputParameters,
        output_schema_type=OutputParameters
    )

    args = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(args.pop("log_level"))

    if 'slice_transform_dict' in args:
        slice_transform = aff.AffineTransform.from_dict(
            args['slice_transform_dict'])
    elif 'slice_transform_list' in args:
        slice_transform = aff.AffineTransform.from_list(
            args['slice_transform_list'])
    else:
        raise ValueError('must provide either an slice_transform_dict '
                         'or an slice_transform_list')

    morphology = morphology_from_swc(args['swc_path'])
    soma_marker = read_soma_marker(args['marker_path'])
    ccf_soma_location = dict(zip(['x', 'y', 'z'], args["ccf_soma_location"]))

    (tilt_correction, tilt_transform) = run_tilt_correction(
        morphology,
        soma_marker,
        ccf_soma_location,
        slice_transform,
        args["slice_image_flip"],
        args['ccf_path'],
    )
    output = {
        'tilt_transform_dict': tilt_transform.to_dict(),
        'tilt_correction': tilt_correction,
        'inputs': parser.args
    }

    parser.output(output)


if __name__ == "__main__":
    main()



""" Utilities (mainly the FeatureWriter class) used by the feature extractor 
executable to format and write outputs.
"""

import os
import copy as cp
import logging
import warnings

from typing import (
    Optional, Any, Dict, List, Iterable, Callable, NamedTuple, Union
)

import h5py
import pandas as pd
import numpy as np

from neuron_morphology.features.layer.layer_histogram import LayerHistogram
from neuron_morphology.feature_extractor.utilities import unnest
from neuron_morphology.features.layer.layer_histogram import \
    EarthMoversDistanceResult


class FeatureWriter:

    def __init__(
        self, 
        heavy_path: str, 
        table_path: Optional[str] = None, 
        formatters: Optional[Iterable["FeatureFormatter"]] = None,
        filemode: Optional[str] = 'w'
    ):
        """ Formats and writes feature extraction outputs

        Parameters
        ----------
        heavy_path : if "heavy" features (e.g. with array outputs) are 
            calculated, write them here.
        table_path : if an output table is requested, write it here

        """

        self.heavy_path = heavy_path
        self.table_path = table_path

        if formatters is None:
            formatters = []


        self.formatters: List["FeatureFormatter"] = list(formatters)
        self.has_heavy = False
        self.output: Dict[str, Any] = {}

        self.validate_table_extension()
        self.heavy_file = h5py.File(self.heavy_path, filemode, driver="core")


    def add_run(self, identifier: str, run: Dict[str, Any]):
        """ Add the results of a feature extraction run to this writer

        Parameters
        ----------
        identifier : the unique identifier for this run
        run : will be added

        """

        run = cp.deepcopy(run)
        features = unnest(run["results"])

        for key in list(features.keys()):
            features[key] = self.process_feature(identifier, key, features[key])

        run["results"] = features

        self.output[identifier] = run 

    def process_feature(self, owner: str, key: str, value: Any):
        """ Processes a feature for writing. This may involve:
            1. changing its type to something json serializable
            2. adding its value to this writer's heavy output

        Parameters
        ----------
        owner : identifies the reconstruction that owns this feature
        key : the name of the feature
        value : the feature's raw value

        Returns
        -------
        Transformed feature value

        """

        matched_handler = None
        matched_name = None

        for name, check, handler in self.formatters:
            if check(key, value):
                if matched_handler is None:
                    matched_handler = handler

                else:
                    warnings.warn(
                        f"skipping formatter: {name} "
                        f"for feature {key} calculated on {owner}"
                        f"({matched_name} matched first)"
                    )
        
        if matched_handler is not None:
            value = matched_handler(self, owner, key, value)

        return value

    def write(self):
        """ Write this writer's output to disk

        Returns
        -------
        This writer's outputs as a dictionary

        """

        if self.has_heavy:
            self.heavy_file.close()

        if self.table_path is not None:
            self.write_table()

        return self.output

    def validate_table_extension(self):
        """ If an output table was requested, check that the path has a 
        supported extension.
        """

        if self.table_path is None:
            return

        self.table_extension = os.path.splitext(self.table_path)[1]
        if self.table_extension not in {".csv"}:
            raise ValueError(f"unsupported extension: {self.table_extension}")

    def build_output_table(self) -> pd.DataFrame:
        """ Convert this writer's output to a reconstruction X feature table

        Returns
        -------
        the generated table

        """

        _table = []
        for reconstruction_id, data in self.output.items():
            current = cp.deepcopy(data["results"])
            current["reconstruction_id"] = reconstruction_id

            # this catches dicts that might have been produced during feature
            # processing
            _table.append(unnest(current))
        
        table = pd.DataFrame(_table)
        table.set_index("reconstruction_id", inplace=True)
        return table

    def write_table(self):
        """ Construct and write a reconstructions X features table from this 
        writer.
        """

        table = self.build_output_table()

        if self.table_extension == ".csv":
            logging.warning(
                "writing additional outputs to csv. See output json for "
                "record of selected features and marks"
            )
            table.to_csv(self.table_path)

        else:
            # just being defensive here - we validate on construction
            raise ValueError(f"invalid extension: {self.table_extension}")

    def register_formatters(self, formatters: Iterable["FeatureFormatter"]):
        """ Add formatters to this writer. The order matters! If multiple 
        formatters match a feature, only the first will be applied.

        Parameters
        ----------
        foratters : an ordered collection of formatter to register

        """
        self.formatters.extend(list(formatters))


# owner, key, value, heavy data store -> transformed value for json
FeatureOutputHandler = Callable[[FeatureWriter, str, str, Any], Any]

# key, value -> is this feature appropriate for this outputter?
FeatureOutputCheck = Callable[[str, Any], bool]

class FeatureFormatter(NamedTuple):
    """ Format feature results for output
    """

    # identifier, used for warning messages mainly
    name: str

    # is this formatter applicable?
    check: FeatureOutputCheck

    # actually format the output
    handler: FeatureOutputHandler

def has_subkey(subkey: str, key: str):
    """ Check whether a string occurs as one of the "."-separated members of 
    another. 
    """

    return subkey in key.split(".")

def add_layer_histogram(
    writer: FeatureWriter,
    owner: str, 
    key: str, 
    histogram: LayerHistogram
) -> str:
    """ Add a layer histogram to this writer's heavy data

    Parameters
    ----------
    owner : identifies the reconstruction that owns this feature
    key : the name of the histogram
    histogram : the histogram's data

    Returns
    -------
    the path at which this writer's heavy data will be stored

    """

    writer.has_heavy = True

    group = writer.heavy_file.create_group(f"{owner}/{key}")
    group.create_dataset("counts", data=histogram.counts)
    group.create_dataset("bin_edges", data=histogram.bin_edges)

    return writer.heavy_path


def process_earth_movers_distance(
    _writer: FeatureWriter,
    _owner: str,
    _key: str,
    value: EarthMoversDistanceResult
) -> Dict[str, Union[str, float]]:
    """ Convert an EarthMoversDistanceResult to a form suitable for json 
    serialization

    Parameters
    ----------
    _* : these are unused
    value : The result to be transformed

    Returns
    -------
    a dictionary like:
    {
        "result": <float result>,
        "interpretation" : <name of interpretation enum value>
    }

    """

    _value = value._asdict()
    _value["interpretation"] = str(_value["interpretation"]).split(".")[1]

    return _value

# ensure small numpy arrays can be serialized to json 
numpy_array_formatter = FeatureFormatter(
    "numpy_array_formatter",
    lambda _, feature: isinstance(feature, np.ndarray),
    lambda _, __, ___, value: value.tolist()
)

# ensure normalized depth histograms written to heavy data
normalized_depth_histogram_formatter = FeatureFormatter(
    "normalized_depth_histogram_formatter",
    lambda key, _: has_subkey("normalized_depth_histogram", key),
    add_layer_histogram
)

# convert earth movers distance results to dicts for json output
earth_movers_distance_formatter = FeatureFormatter(
    "earth_movers_distance_formatter",
    lambda key, _: has_subkey("earth_movers_distance", key),
    process_earth_movers_distance
)

DEFAULT_FEATURE_FORMATTERS = (
    normalized_depth_histogram_formatter,
    earth_movers_distance_formatter,
    numpy_array_formatter,

)



""" Some handy utilities for working with vector geometries
"""

from typing import List, Tuple, Callable
import warnings
from shapely import geometry as geo


def get_ccw_vertices_from_two_lines(line1: List[Tuple], line2: List[Tuple]):
    """
        Convenience method two do both get_vertices_from_two_lines()
        and get_ccw_vertices()
    """
    return get_ccw_vertices(get_vertices_from_two_lines(line1, line2))

def prune_two_lines(line1: List[Tuple], line2: List[Tuple]):
    """
        check the boundary to avoid intersections with side lines

        Parameters
            ----------
            line1, line2: List of coordinates describing two lines

            Returns
            -------
            line1, line2: boundary pruned if needed
    """
    # validate the pia/wm does not cross the side lines
    prune = True
    while prune:
        prune = False

        # update lines
        line1_str = geo.LineString(line1)
        line2_str = geo.LineString(line2)

        side1 = geo.LineString([line1[0], line2[-1]])
        side2 = geo.LineString([line1[-1], line2[0]])

        # prune the edge points if there are intersections of side and pia/wm
        if side1.crosses(line1_str):
            line1.pop(0)
            prune = True

        if side1.crosses(line2_str):
            line2.pop(-1)
            prune = True

        if side2.crosses(line1_str):
            line1.pop(-1)
            prune = True
        
        if side2.crosses(line2_str):
            line2.pop(0)
            prune = True

        if prune:
            warnings.warn(f"lines are modified \nline1: {line1}\nline2: {line2}", UserWarning)

    return line1, line2

def get_vertices_from_two_lines(line1: List[Tuple], line2: List[Tuple]):
    """
        Generates circular vertices from two lines

        Parameters
        ----------
        line1, line2: List of coordinates describing two lines

        Returns
        -------
        vertices of the simple polygon created from line 1 and 2
        (first vertex = last vertex)

        1-2-3-4
        5-6-7-8 -> [1-2-3-4-8-7-6-5-1]

    """
    side1 = geo.LineString([line1[0], line2[-1]])
    side2 = geo.LineString([line1[-1], line2[0]])

    if side1.crosses(side2):
        line2.reverse()

    line1, line2 = prune_two_lines(line1, line2)

    vertices = line1 + line2 + [line1[0]]
    return vertices

def get_ccw_vertices(vertices: List[Tuple]):
    """
        Generates counter clockwise vertices from vertices describing
        a simple polygon

        Method: Simplification of the shoelace formula, which calculates
        area of a simple polygon by integrating the area under each line
        segment of the polygon. If the total area is positive, the vertices
        were traversed in clockwise order, and if it is negative, they were
        traversed in counterclockwise order.

        Parameters
        ----------
        vertices: vertices describing a convex polygon
                  (vertices[0] = vertices[-1])

        Returns
        -------
        vertices in counter clockwise order

    """

    winding = 0
    for i in range(len(vertices)-1):
        (x0, y0) = vertices[i]
        (x1, y1) = vertices[i+1]
        winding += (x1 - x0) * (y1 + y0)

    if winding > 0:
        vertices.reverse()

    return vertices


from typing import (
    AbstractSet, Set, Collection, Optional, Dict, Type, FrozenSet, List)
import logging
import warnings

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.marked_feature import MarkedFeature
from neuron_morphology.feature_extractor.mark import Mark


class FeatureExtractionRun:
    
    def __init__(self, data):
        """ Represents a single run of feature extraction on a single dataset.

        Parameters
        ----------
        data : the dataset from which to extract features

        """

        self.data: Data = data

        self.selected_marks: Set[Type[Mark]] = set()
        self.selected_features: List[MarkedFeature] = []
        self.results: Optional[Dict] = None

    def select_marks(
        self, 
        marks: Collection[Type[Mark]], 
        required_marks: AbstractSet[Type[Mark]] = frozenset()
    ):
        """ Choose marks for this run by validating a set of candidates 
        against the data.

        Parameters
        ----------
        marks : candidate marks to be validated
        required_marks : if provided, raise an exception if any of these marks
            do not validate successfully

        Returns
        -------
        self : This FeatureExtractionRun, with selected_marks updated

        """

        for mark in marks:
            if mark.validate(self.data):
                self.selected_marks.add(mark)
            else:
                logging.info(f"skipping mark (validation failed): {mark.__class__.__name__}")

        missing_required = required_marks - self.selected_marks
        if missing_required:
            raise ValueError(f"required marks: {missing_required} failed validation!")

        logging.info(f"selected marks: {self.select_marks}")
        return self

    def select_features(
        self, 
        features: Collection[MarkedFeature],
        only_marks: Optional[AbstractSet[Type[Mark]]] = None, 

    ):
        """ Choose features to calculated for this run on the basis of selected
        marks.

        Parameters
        ----------
        features : Candidates features for selection
        only_marks : if provided, reject features not marked with marks in 
            this set

        Returns
        -------
        self : This FeatureExtractionRun, with selected_features updated

        """
        if only_marks is None:
            only_marks = set()

        for feature in features:
            extra_marks = feature.marks - self.selected_marks
            if extra_marks:
                logging.info(
                    f"skipping feature: {feature.name}. "
                    f"Found extra marks: {[mark.__name__ for mark in extra_marks]}")
            elif only_marks - feature.marks:
                logging.info(f"skipping feature: {feature.name} (no marks from {only_marks})")
            else:
                self.selected_features.append(feature)

        
        logging.info(f"selected features: {[feature.name for feature in self.selected_features]}")
        return self

    def extract(self):
        """ For each selected feature, carry out calculation on this run's 
        dataset.

        Returns
        -------
        self : This FeatureExtractionRun, with results updated
        """

        self.results = {}

        for feature in self.selected_features:
            try:
                self.results[feature.name] = feature(self.data)
            except:
                logging.warning(f"feature extraction failed for {feature.name}")
                raise

        return self

    def serialize(self):
        """ Return a dictionary describing this run
        """

        return {
            "results": self.results,
            "selected_marks": [mark.__name__ for mark in self.selected_marks],
            "selected_features": [
                feature.name for feature in self.selected_features]
        }



from unittest import TestCase, skipIf
import subprocess as sp
import os
import tempfile
import shutil
import json

from dictdiffer import diff


class SnapPolygonsIntegration(TestCase):

    @skipIf(
        os.environ.get("TEST_INTEGRATION", "false") != "true",
        "The integration tests can take a while - must opt in!"
    )
    def test_integration(self):
        data_root = os.path.join(os.path.dirname(__file__), "data")
        cases = [
            ["678492000_inputs.json", "678492000_outputs.json"],
            ["680352737_inputs.json", "680352737_outputs.json"],
            ["735355641_inputs.json", "735355641_outputs.json"],
            ["839050458_inputs.json", "839050458_outputs.json"]
        ]

        tmpdir = tempfile.mkdtemp()

        for in_fname, out_fname in cases:
            with self.subTest():
                full_in_json_path = os.path.join(data_root, in_fname)
                full_out_json_path = os.path.join(tmpdir, out_fname)
                full_expected_path = os.path.join(data_root, out_fname)

                sp.check_call([
                    "python", "-m", "neuron_morphology.snap_polygons",
                    "--input_json", full_in_json_path,
                    "--output_json", full_out_json_path
                ])

                with open(full_out_json_path, "r") as out_json:
                    obtained = json.load(out_json)
                with open(full_expected_path, "r") as expected_json:
                    expected = json.load(expected_json)

                difference = diff(expected, obtained)
                self.assertEqual(list(difference), [])

        shutil.rmtree(tmpdir)



import os
import boto3
import json
import configparser
from zipfile import ZipFile, ZIP_DEFLATED
from io import BytesIO
from argschema import ArgSchemaParser
from neuron_morphology.pipeline._schemas import InputParameters

def get_credentials(credentials_file):
    """
    get credentisals from credentials_file

    Parameters
    ---------------
    credentials_file: file path to credentials_file

    Return
    ---------------
    credentials: access_key_id, secret_access_key

    """
    config = configparser.ConfigParser()
    config.read(credentials_file)
    access_key_id = None
    secret_access_key = None
    if 'default' in config.sections():
        if 'aws_access_key_id' in config.options('default'):
            access_key_id = config.get('default', 'aws_access_key_id')
        if 'aws_access_key_id' in config.options('default'):
            secret_access_key = config.get('default', 'aws_secret_access_key')

    return access_key_id, secret_access_key

def zip_files(file_dict):
    """
    zip files into an archive in memory

    Parameters
    ---------------
    file_dict: file name: file paths or file in bytes to be archived

    Return
    ---------------
    archive: BytesIO obj

    """

    archive = BytesIO()

    with ZipFile(archive, "a", ZIP_DEFLATED, False) as zip_file:
        for name, file in file_dict.items():
            if isinstance(file, str):
                with open(file, 'rb') as fh:
                    data = BytesIO(fh.read())
                zip_file.writestr(name, data.getvalue())
            elif isinstance(file, BytesIO):
                zip_file.writestr(name, file.getvalue())
            else:
                raise TypeError("Invalid input file!")

    return archive


def post_object_to_s3(data, name, bucket, region, access_key_id=None, secret_access_key=None):
    """
    post object (bytes) in memory to S3 bucket

    Parameters
    ------------------
    data: the object data
    name: the object data's name in s3
    region: where the s3 bucket located
    bucket: s3 bucket name or arn
    access_key_id, secret_access_key: aws user's credentials
        implicitly the credentials file located in ~/.aws/credentials or
        set AWS_SHARED_CREDENTIALS_FILE to the credentials file in your environment
    
    Return
    ------------------
    True if successful
    
    """
    if access_key_id and secret_access_key:
        s3_client = boto3.client(
            's3',
            aws_access_key_id=access_key_id,
            aws_secret_access_key=secret_access_key,
            region_name=region
        )
    else:
        s3_client = boto3.client('s3', region_name=region)

    response = s3_client.put_object(Bucket=bucket, 
                                Key=name, 
                                Body=data)


def main():
    """
    Usage:
    python post_data_to_s3.py --input_json INPUT_JSON
    """
    parser = ArgSchemaParser(schema_type=InputParameters)
    inputs = parser.args

    file_dict = {}

    swc_file_name = os.path.basename(inputs['swc_file'])
    file_dict[swc_file_name] = inputs['swc_file']
    inputs['swc_file'] = swc_file_name

    if inputs['marker_file'] is not None:
        marker_file_name = os.path.basename(inputs['marker_file'])
        file_dict[marker_file_name] = inputs['marker_file']
        inputs['marker_file'] = marker_file_name

    json_data = json.dumps(inputs)
    binary_data = json_data.encode()
    input_json = BytesIO(binary_data)

    json_fn = str(inputs['neuron_reconstruction_id']) + ".json"
    file_dict[json_fn] = input_json

    archive_data = zip_files(file_dict)
    archive_name = str(inputs['neuron_reconstruction_id']) + ".zip"
    bucket = inputs['destination_bucket']['name']
    bucket_region = inputs['destination_bucket']['region']
    credentials_file = inputs['destination_bucket']['credentials_file']
    access_key_id, secret_access_key = get_credentials(credentials_file)

    post_object_to_s3(archive_data.getvalue(), archive_name, bucket, bucket_region, access_key_id, secret_access_key)


if __name__ == "__main__":
    main()


import unittest

from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.constants import (
    SOMA, AXON, APICAL_DENDRITE, BASAL_DENDRITE)


class TestMorphologyBuilder(unittest.TestCase):

    def test_branching(self):

        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .axon(0, 0, 1)
                        .axon(0, 0, 2)
                            .axon(0, 0, 3).up()
                            .axon(0, 0, 4)
                                .axon(0, 0, 5).up()
                                .axon(0, 0, 6).up(4)
                    .basal_dendrite()
                        .basal_dendrite()
                .build()
        )

        obtained = morphology.get_branching_nodes()
        self.assertEqual(obtained[0]["z"], 2)
        self.assertEqual(obtained[1]["z"], 4)
        self.assertEqual(len(obtained), 2)

    def test_multiroot(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .axon(0, 0, 1)
                .root(1, 1, 1, AXON)
                    .child(2, 1, 1, AXON)
                .build()
        )

        roots = morphology.get_roots()
        self.assertEqual(roots[1]["parent"], -1)
        self.assertEqual(morphology.nodes()[-1]["parent"], 2)


from typing import Optional, List, Dict

import numpy as np

from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.mark import (
    RequiresRoot,
    BifurcationFeatures,
    Geometric,
)
from neuron_morphology.feature_extractor.data import (
    MorphologyLike, get_morphology)
from neuron_morphology.morphology import Morphology


__all__ = [
    "num_outer_bifurcations",
    "calculate_outer_bifs",
    "mean_bifurcation_angle_local",
    "mean_bifurcation_angle_remote"
]


def calculate_outer_bifs(
    morphology: Morphology,
    soma: Dict,
    node_types: Optional[List[int]]
) -> int:

    """
        Counts the number of bifurcation points beyond the a sphere
        with 1/2 the radius from the soma to the most distant point
        in the morphology, with that sphere centered at the soma.

        Parameters
        ----------
        morphology: Describes the structure of a neuron
        soma: Must have keys "x", "y", and "z", describing the position of this
            morphology's soma in
        node_types: Restrict included nodes to these types. See
            neuron_morphology.constants for avaiable node types.

        Returns
        -------
        the number of bifurcations

    """

    nodes = morphology.get_node_by_types(node_types)
    far = 0
    for node in nodes:
        dist = morphology.euclidean_distance(soma, node)
        if dist > far:
            far = dist

    count = 0
    rad = far / 2.0
    for node in nodes:
        if len(morphology.children_of(node)) > 1:
            dist = morphology.euclidean_distance(soma, node)
            if dist > rad:
                count += 1
    return count


@marked(Geometric)
@marked(BifurcationFeatures)
@marked(RequiresRoot)
def num_outer_bifurcations(
    data: MorphologyLike,
    node_types: Optional[List[int]] = None
) -> int:
    """ Feature Extractor interface to calculate_outer_bifurcations. Returns
    the number of bifurcations (branch points), excluding those too close to
    the root (threshold is 1/2 the max distance from the root to any node).

    Parameters
    ----------
    data : Holds a morphology object. No additional data is required
    node_types : Restrict included nodes to these types. See
        neuron_morphology.constants for avaiable node types.

    """

    return calculate_outer_bifs(
        data.morphology,
        data.morphology.get_root(),
        node_types
    )


def angle_between(v1, v2):
    """Helper function to get angle between two numpy vectors"""
    v1 = np.asarray(v1)
    v2 = np.asarray(v2)
    v1_u = v1 / np.linalg.norm(v1)
    v2_u = v2 / np.linalg.norm(v2)

    return np.arccos(np.clip(np.dot(v1_u, v2_u), -1.0, 1.0))


@marked(Geometric)
@marked(BifurcationFeatures)
def mean_bifurcation_angle_local(
    data: MorphologyLike,
    node_types: Optional[List[int]] = None
) -> float:
    """
        Compute the average angle between child segments at
        bifurcations throughout the morphology.
        Trifurcations are ignored. Note: this introduces possible segmentation
        artifacts if trifurcations are due to large segment sizes.

        Parameters
        ----------
        data: The reconstruction whose max euclidean distance will be
            calculated
        node_types: restrict consideration to these types

        Returns
        -------

        Scalar value

    """
    morphology = get_morphology(data)
    total_angle = 0.0
    n = 0
    nodes = morphology.get_node_by_types(node_types)
    for node in nodes:
        if len(morphology.children_of(node)) == 2:
            node_vec = np.asarray([node['x'], node['y'], node['z']])

            a = morphology.children_of(node)[0]
            a_vec = np.asarray([a['x'], a['y'], a['z']])

            b = morphology.children_of(node)[1]
            b_vec = np.asarray([b['x'], b['y'], b['z']])

            total_angle += angle_between(a_vec - node_vec, b_vec - node_vec)
            n += 1

    if n == 0:
        return float('nan')
    return total_angle / n


@marked(Geometric)
@marked(BifurcationFeatures)
def mean_bifurcation_angle_remote(
    data: MorphologyLike,
    node_types: Optional[List[int]] = None
) -> float:
    """
        Compute the average angle between the next branch point or terminal
        tip of child segments at each bifurcation.
        Trifurcations are ignored. Note: this introduces possible segmentation
        artifacts if trifurcations are due to large segment sizes.

        Parameters
        ----------

        Parameters
        ----------
        data: The reconstruction whose max euclidean distance will be
            calculated
        node_types: restrict consideration to these types


        Returns
        -------

        Scalar value, nan if no nodes

    """
    morphology = get_morphology(data)
    total_angle = 0.0
    n = 0
    nodes = morphology.get_node_by_types(node_types)
    for node in nodes:
        if len(morphology.children_of(node)) == 2:
            node_vec = np.asarray([node['x'], node['y'], node['z']])

            # find the point to measure to, whether it be the next
            #   branch point or tip
            a = morphology.children_of(node)[0]
            while len(morphology.children_of(a)) == 1:
                a = morphology.children_of(a)[0]
            a_vec = np.asarray([a['x'], a['y'], a['z']])

            b = morphology.children_of(node)[1]
            while len(morphology.children_of(b)) == 1:
                b = morphology.children_of(b)[0]
            b_vec = np.asarray([b['x'], b['y'], b['z']])

            total_angle += angle_between(a_vec - node_vec, b_vec - node_vec)
            n += 1

    if n == 0:
        return float('nan')
    return total_angle / n



from tests.objects import test_node, test_tree
from neuron_morphology.validation.result import InvalidMorphology
from neuron_morphology.validation import type_validation as tv
from tests.validation.validation_test_case import ValidationTestCase
from neuron_morphology.constants import *
import unittest
from mock import patch


class TestTypeValidationFunctions(ValidationTestCase):
    """ Tests the functions in type_validation.py """

    def test_validate_expected_type_valid(self):
        for node_type in [AXON, BASAL_DENDRITE, APICAL_DENDRITE]:
            nodes = [test_node(id=1, type=SOMA), test_node(id=2, type=node_type, parent_node_id=1)]
            errors = tv.validate_expected_types(test_tree(nodes))
            self.assertEqual(len(errors), 0)

    def test_validate_expected_type_invalid(self):
        nodes = [test_node(type=5)]
        errors = tv.validate_expected_types(test_tree(nodes))
        self.assertNodeErrors(errors, "Node type needs to be one of these values:", [[1]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_soma_node_with_valid_parent(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1)]
        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_soma_node_with_invalid_parent(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=SOMA, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 1 can only have a parent of the following types:", [[2]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_axon_node_with_valid_parent_type(self):
        for node_type in [SOMA, AXON]:
            nodes = [test_node(id=1, type=node_type, parent_node_id=-1), test_node(id=2, type=AXON, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_axon_node_with_invalid_parent_type(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1),
                     test_node(id=3, type=AXON, parent_node_id=2)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 2 can only have a parent of the following types:", [[3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_basal_dendrite_node_with_valid_parent_type(self):
        for node_type in [SOMA]:
            nodes = [test_node(id=1, type=node_type, parent_node_id=-1),
                     test_node(id=2, type=BASAL_DENDRITE, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_basal_dendrite_node_with_invalid_parent_type(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=AXON, parent_node_id=1),
                     test_node(id=3, type=BASAL_DENDRITE, parent_node_id=2)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 3 can only have a parent of the following types:", [[3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_basal_dendrite_node_with_invalid_parent_type_independent(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=AXON, parent_node_id=1),
                     test_node(id=3, type=BASAL_DENDRITE, parent_node_id=-1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 3 can only have a parent of the following types:", [[3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_apical_dendrite_node_with_invalid_parent_type_independent(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=AXON, parent_node_id=1),
                     test_node(id=3, type=APICAL_DENDRITE, parent_node_id=-1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 4 can only have a parent of the following types:", [[3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_apical_dendrite_node_with_valid_parent_type(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1)]
        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_apical_node_with_invalid_parent_type(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=AXON, parent_node_id=1),
                     test_node(id=3, type=BASAL_DENDRITE, parent_node_id=1),
                     test_node(id=4, type=APICAL_DENDRITE, parent_node_id=3)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Type 4 can only have a parent of the following types:", [[4]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_number_of_soma_nodes_invalid(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1), test_node(id=2, type=SOMA, parent_node_id=-1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "The morphology needs to have one soma node", [[1], [2]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_number_of_apical_dendrite_with_parent_of_soma_invalid(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1),
                     test_node(id=3, type=APICAL_DENDRITE, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Nodes of type 4 can only have 1 parent of type 1", [[2], [3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_immediate_child_of_soma_doesnt_more_than_one_child(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1),
                 test_node(id=3, type=APICAL_DENDRITE, parent_node_id=2),
                 test_node(id=4, type=APICAL_DENDRITE, parent_node_id=3),
                 test_node(id=5, type=APICAL_DENDRITE, parent_node_id=3)]
        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_immediate_child_of_soma_has_more_than_one_child(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=APICAL_DENDRITE, parent_node_id=1),
                     test_node(id=3, type=APICAL_DENDRITE, parent_node_id=2),
                     test_node(id=4, type=APICAL_DENDRITE, parent_node_id=2)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Immediate children of soma cannot branch", [[2]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_multiple_axon_initiation_points_not_valid(self):
        try:
            nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                     test_node(id=2, type=AXON, parent_node_id=1),
                     test_node(id=3, type=AXON, parent_node_id=1)]
            test_tree(nodes, strict_validation=True)
            self.fail("Morphology should have been rejected.")
        except InvalidMorphology as e:
            self.assertNodeErrors(e.validation_errors, "Axon can only have one parent of type basal dendrite or soma",
                                  [[2], [3]])

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_multiple_axon_initiation_points_valid(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=AXON, parent_node_id=1)]

        test_tree(nodes, strict_validation=True)

    @patch("neuron_morphology.validation.swc_validators", [tv])
    def test_multiple_axon_initiation_points_valid_with_independent_axon(self):
        nodes = [test_node(id=1, type=SOMA, parent_node_id=-1),
                 test_node(id=2, type=AXON, parent_node_id=1),
                 test_node(id=3, type=AXON, parent_node_id=-1)]

        test_tree(nodes, strict_validation=True)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTypeValidationFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)



"""Utilites for writing diagnostic overlay images
"""
import os
import itertools as it
from typing import Sequence, Tuple, Dict, Optional
import copy as cp

import glymur
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
import imageio

from neuron_morphology.snap_polygons.geometries import Geometries, make_scale


class ImageOutputter:
    """ Overlays polygons and surfaces on provided images. Writes the 
    results to files.

    Parameters
    ----------
    native_geo : Layer geometries before gaps are filled
    result_geo : Layer geometries after gaps are filled
    image_specs : Each is a dictionary defining a single image. Must 
        provide string keys:
            - input_path : read from here
            - output_path : write to (siblings of) this path
            - downsample : the image will be scaled by this factor in each 
                dimension
            - overlay_types : produce these kinds of overlay for this image
    alpha : of the transparent overlays
    color_cycle : as polygon fills are drawn, cycle through these colors
    savefig_kwargs : Passed directly to pyplot's savefig, use to specify 
        e.g dpi.
    """

    DEFAULT_COLOR_CYCLE = ("c", "m", "y", "k", "r", "g", "b")
    OVERLAY_TYPES = {
        "before": "draw_before",
        "after": "draw_after"
    }

    def __init__(
            self, 
            native_geo: Geometries, 
            result_geo: Geometries, 
            image_specs: Optional[Sequence[Dict]],
            alpha: float = 0.4,
            color_cycle: Optional[Sequence] = None,
            savefig_kwargs: Optional[Dict] = None
    ):

        if color_cycle is None:
            color_cycle = ImageOutputter.DEFAULT_COLOR_CYCLE

        self.native_geo = native_geo
        self.result_geo = result_geo
        self.image_specs = image_specs or []

        self.alpha = alpha
        self.color_cycle = color_cycle
        self.overlay_types = cp.copy(ImageOutputter.OVERLAY_TYPES)

        _savefig_kwargs = {"dpi": 300}
        if savefig_kwargs is not None:
            _savefig_kwargs.update(savefig_kwargs)
        self.savefig_kwargs = _savefig_kwargs

    def _draw_geometries(
            self, 
            geometries: Geometries, 
            image: np.ndarray
    ):
        """ Utility for overlaying polygons and surfaces on an image. See 
        draw_before and draw_after for more details.
        """

        fig, axes = plt.subplots()
        axes.imshow(image)

        cycler = it.cycle(self.color_cycle)
        for color, (name, poly) in zip(cycler, geometries.polygons.items()):
            patch = make_pathpatch(
                poly.exterior.coords, 
                alpha=self.alpha, 
                lw=0.25, 
                label=name, 
                facecolor=color, 
                edgecolor="k"
            )
            axes.add_patch(patch)
        
        for name, surf in geometries.surfaces.items():
            patch = make_pathpatch(
                surf.coords, fill=False, lw=0.25, label=name
            )
            axes.add_patch(patch)

        axes.set_axis_off()
        return fig

    def draw_before(
            self, 
            image: np.ndarray,
            scale: float = 1.0
    ):
        """ Display the pre-fill polygons and surfaces overlaid on an image.

        Parameters
        ----------
        image : onto which objects will be drawn
        scale : required to transform from object space to image space

        Returns
        -------
        A matplotlib figure containing the overlay

        """

        return self._draw_geometries(
            self.native_geo.transform(make_scale(scale)),
            image
        )

    def draw_after(
            self, 
            image: np.ndarray,
            scale: float = 1.0
    ):
        """ Display the post-fill polygons and surfaces overlaid on an image.

        Parameters
        ----------
        image : onto which objects will be drawn
        scale : required to transform from object space to image space

        Returns
        -------
        A matplotlib figure containing the overlay

        """

        return self._draw_geometries(
            self.result_geo.transform(make_scale(scale)),
            image
        )

    def write_images(self):
        """ For each image specified in this outputter and each overlay type
        requested for that image, produce and save an overlay.
        """

        written = []

        for image_spec in self.image_specs:
            image = read_image(
                image_spec["input_path"], 
                image_spec["downsample"]
            )

            for overlay_type in image_spec["overlay_types"]:
                if overlay_type not in self.overlay_types:
                    raise ValueError(
                        f"unrecognized overlay type: {overlay_type} "
                        f"(options: {list(self.overlay_types.keys())})"
                    )

                fig = getattr(
                    self, self.overlay_types[overlay_type]
                )(image, 1.0 / image_spec["downsample"])

                output_path = fname_suffix(
                    image_spec["output_path"], overlay_type)
                write_figure(fig, output_path, **self.savefig_kwargs)
    
                written.append({
                    "input_path": image_spec["input_path"],
                    "downsample": image_spec["downsample"],
                    "output_path": output_path,
                    "overlay_type": overlay_type
                })

        return written


def write_figure(fig: plt.Figure, *args, **kwargs):
    """ Write a matplotlib figure without respect to the current figure.

    Parameters
    ----------
    fig : the figure to be writter    
    *args, **kwargs : passed to plt.savefig

    """

    cf_number = plt.gcf().number
    plt.figure(fig.number)
    plt.savefig(*args, **kwargs)
    plt.figure(cf_number)


def read_image(path: str, decimate: int = 1):
    """ Read an image. Dispatch to an appropriate library based on that 
    image's extension.

    Parameters
    ----------
    path : to the image
    decimate : apply a decimation of this factor along each axis of the image

    """

    _, ext = os.path.splitext(path)

    if ext == ".jp2":
        return read_jp2(path, decimate)
    else:
        return read_with_ndimage(path, decimate)


def read_with_ndimage(path: str, decimate: int):
    """Read (and symmetrically decimate) an image file into a numpy array
    """
    return imageio.imread(path)[::decimate, ::decimate, ...]


def read_jp2(path: str, decimate: int):
    """Read (and symmetrically decimate) a jp2 file into a numpy array 
    """
    fil = glymur.Jp2k(path)
    return fil[::decimate, ::decimate]


def fname_suffix(path: str, suffix: str):
    """ Utility for adding a suffix to a path string. The suffix will be 
    inserted before the extension.
    """

    head, ext = os.path.splitext(path)
    if ext == "":
        return f"{head}_{suffix}"
    else:
        return f"{head}_{suffix}{ext}"

    
def make_pathpatch(
        vertices: Sequence[Tuple[float, float]], 
        **patch_kwargs
) -> mpl.patches.PathPatch:
    """ Utility for building a matplotlib pathpatch from an array of vertices

    Parameters
    ----------
    vertices : Defines the path. May be closed or open
    **patch_kwargs : passed directly to pathpatch constructor

    """

    codes = [mpl.path.Path.MOVETO] \
        + [mpl.path.Path.LINETO] * (len(vertices) - 1)
    path = mpl.path.Path(vertices, codes)
    return mpl.patches.PathPatch(path, **patch_kwargs)



from typing import Dict, Optional, List, Any, Union

import logging
import copy as cp

import numpy as np
import pandas as pd
import warnings
from argschema.argschema_parser import ArgSchemaParser
from neuron_morphology.transforms.scale_correction._schemas import (
    InputParameters, OutputParameters)
from neuron_morphology.morphology import Morphology
from neuron_morphology.swc_io import morphology_from_swc
import neuron_morphology.transforms.affine_transform as aff


def estimate_scale_correction(morphology: Morphology,
                              soma_depth: float,
                              soma_marker_z: float,
                              cut_thickness: Optional[float] = 350):
    """
        Estimate a scale factor to correct the reconstructed morphology
        for slice shrinkage

        Prior to reconstruction, the slice shrinks due to evaporation.
        This is most notable in the z axis, which is the slice thickness.

        To correct for shrinkage we compare soma depth within the slice
        obtained soon after cutting the slice to the fixed_soma_depth obtained
        during the reconstruction. Then the scale correction is estimated as:
        scale  = soma_depth / fixed_soma_depth.
        This is sensible as long as the z span of the corrected reconstruction
        is contained  within the slice thickness. Thus we also estimate
        the maximum scale correction as:
        scale_max  = cut_thickness / z_span,
        and take the smaller of scale and scale_max


        Parameters
        ----------
        morphology: Morphology object
        soma_depth: recorded depth of the soma when it was sliced
        soma_marker_z: soma marker z value from revised marker file
                       (z is on the slice surface for the marker file)
        cut_thickness: thickness of the cut slice

        Returns
        -------
        scale factor correction

    """
    soma_morph_z = morphology.get_soma()['z']
    fixed_depth = np.abs(soma_morph_z - soma_marker_z)
    scale = soma_depth / fixed_depth

    node_z = [node['z'] for node in morphology.nodes()]
    z_range = np.max(node_z) - np.min(node_z)
    scale_max = cut_thickness / z_range

    if scale > scale_max:
        scale = scale_max
        warnings.warn(f"Shrinkage scale correction factor: {scale} "
                      f"exceeded the max allowed value: {scale_max}. "
                      f"Will correct for shrinkage using the maximum value."
                      )

    return scale


def get_soma_marker_from_marker_file(marker_path: str):
    col_names = ['x', 'y', 'z', 'radius', 'shape', 'name',
                 'comment', 'color_r', 'color_g', 'color_b']
    markers = pd.read_csv(marker_path, sep=',', comment='#',
                          header=None, names=col_names)
    soma_marker = markers.loc[markers['name'] == 30].to_dict('r')
    return soma_marker[0]


def run_scale_correction(
        morphology: Morphology,
        soma_marker_z: float,
        soma_depth: float,
        cut_thickness: float):


    scale_correction = estimate_scale_correction(morphology,
                                      soma_depth,
                                      soma_marker_z,
                                      cut_thickness)

    scale_transform = aff.affine_from_transform(
        [[1, 0, 0],
         [0, 1, 0],
         [0, 0, scale_correction]]
    )
    at = aff.AffineTransform(scale_transform)
    morphology_scaled = at.transform_morphology(morphology,
                                                scale_radius=False)

    return {
        "morphology_scaled": morphology_scaled,
        "scale_transform": at.to_dict(),
        "scale_correction": scale_correction,
    }

def collect_inputs(args: Dict[str,Any]) -> Dict[str,Any]:
    """

    Parameters
    ----------
    args: dict of InputParameters

    Returns
    -------
    dict with string keys:
        morphology: Morphology object
        soma_marker_z: z value from the marker file
        soma_depth: soma depth
        cut_thickness: slice thickness
    """
    morphology = morphology_from_swc(args["swc_path"])
    soma_marker = get_soma_marker_from_marker_file(args["marker_path"])
    soma_depth = args["soma_depth"]
    cut_thickness = args["cut_thickness"]

    return {
        "morphology": morphology,
        "soma_marker_z": soma_marker["z"],
        "soma_depth": soma_depth,
        "cut_thickness": cut_thickness,
    }

def main():
    parser = ArgSchemaParser(
        schema_type=InputParameters,
        output_schema_type=OutputParameters
    )

    args = cp.deepcopy(parser.args)
    logging.getLogger().setLevel(args.pop("log_level"))

    inputs = collect_inputs(args)
    outputs = run_scale_correction(**inputs)
    outputs.pop("morphology_scaled")

    outputs.update({"inputs": parser.args})
    parser.output(outputs)


if __name__ == "__main__":
    main()



from typing import Optional, List
from functools import partial

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.features.statistics.coordinates import COORD_TYPE

from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.mark import Intrinsic


@marked(Intrinsic)
def num_tips(
        data: Data,
        node_types: Optional[List] = None,
        ):
    """
        Calculate number of tips

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

    """
    # Alternative method:
    num_tips = len(COORD_TYPE.TIP.get_coordinates(data.morphology,
                                                  node_types=node_types))
    return num_tips


@marked(Intrinsic)
def num_nodes(
        data: Data,
        node_types: Optional[List] = None,
        ):
    """
        Calculate number of nodes of a given type

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

    """
    num_nodes = len(data.morphology.get_node_by_types(node_types))
    return num_nodes


def child_ids_by_type(node_id, morphology, node_types=None):
    """ Helper function for the traversal functions"""
    node = morphology.node_by_id(node_id)
    children = morphology.get_children(
                    node, node_types=node_types)
    return [child['id'] for child in children]


def calculate_branches_from_root(morphology,
                                 root,
                                 node_types=None):
    """
        Calculate the number of branches of a specific neuron type
        in a morphology. A branch is defined as being between
        two bifurcations or between a bifurcation and a tip
        if a node has three or more children, it is treated as succesive
        bifurcations, e.g a trifurcation: _/_/__ creates 4 branches
        since the branch between the two bifurcations counts

        Parameters
        ----------

        morphology: a morphology object

        root: the root node to traverse from

        node_types: a list of node types (see neuron_morphology constants)
    """
    counter = {'num_branches': 0}

    def branch_visitor(node, counter, node_types):
        num_children = len(morphology.get_children(node,
                                                   node_types=node_types))
        if num_children > 1:
            # branches + (implicit branches from successive bifurcations)
            counter['num_branches'] += num_children + (num_children - 2)
        elif counter['num_branches'] == 0:  # still count root with one child
            counter['num_branches'] += 1

    visitor = partial(branch_visitor,
                      counter=counter,
                      node_types=node_types)
    neighbor_cb = partial(child_ids_by_type,
                          morphology=morphology,
                          node_types=node_types)
    morphology.breadth_first_traversal(visitor,
                                       start_id=morphology.node_id_cb(root),
                                       neighbor_cb=neighbor_cb)
    return counter['num_branches']


@marked(Intrinsic)
def num_branches(
        data: Data,
        node_types: Optional[List] = None,
        ):
    """
        Calculate number of branches

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

    """
    morphology = data.morphology
    nodes = morphology.get_node_by_types(node_types)
    roots = morphology.get_roots_for_nodes(nodes)
    num_branches = 0
    for root in roots:
        num_branches += calculate_branches_from_root(
            morphology, root, node_types=node_types)
        if (morphology.parent_of(root) is not None and
            len(morphology.get_children_of_node_by_types(root, node_types)) > 1):
             # if root is a branching node, include the branch that connects root to tree
            num_branches += 1
    return num_branches


def calculate_mean_fragmentation_from_root(morphology,
                                           root,
                                           node_types=None):
    """
        Calculate the mean fragmentation from a root
        in a morphology. Mean fragmentation is the number of compartments
        over the number of branches. A branch is defined as being between
        two bifurcations or between a bifurcation and a tip
        if a node has three or more children, it is treated as succesive
        bifurcations, e.g a trifurcation: _/_/__ creates 4 branches
        since the branch between the two bifurcations counts

        Parameters
        ----------

        morphology: a morphology object

        root: the root node to traverse from

        node_types: a list of node types (see neuron_morphology constants)
    """
    counter = {'num_branches': 0,
               'num_compartments': 0}

    def branch_visitor(node, counter, node_types):
        num_children = len(morphology.get_children(node,
                                                   node_types=node_types))
        if num_children > 1:
            # branches + (implicit branches from successive bifurcations)
            counter['num_branches'] += num_children + (num_children - 2)
            counter['num_compartments'] += num_children + (num_children - 2)
        elif num_children == 1:
            counter['num_compartments'] += 1
            if counter['num_branches'] == 0:  # still count root with one child
                counter['num_branches'] += 1

    visitor = partial(branch_visitor,
                      counter=counter,
                      node_types=node_types)
    neighbor_cb = partial(child_ids_by_type,
                          morphology=morphology,
                          node_types=node_types)
    morphology.breadth_first_traversal(visitor,
                                       start_id=morphology.node_id_cb(root),
                                       neighbor_cb=neighbor_cb)

    mean_fragmentation = counter['num_compartments'] / counter['num_branches']
    return (mean_fragmentation,
            counter['num_branches'],
            counter['num_compartments'])


@marked(Intrinsic)
def mean_fragmentation(
        data: Data,
        node_types: Optional[List] = None,
        ):
    """
        Calculate the mean number of compartments per branch

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

    """
    morphology = data.morphology
    roots = morphology.get_roots()
    num_branches = 0
    num_compartments = 0
    for root in roots:
        (_, local_branches, local_compartments) = \
            calculate_mean_fragmentation_from_root(
                morphology, root, node_types=node_types)

        num_branches += local_branches
        num_compartments += local_compartments

    mean_fragmentation = num_compartments / num_branches
    return mean_fragmentation


def calculate_max_branch_order_from_root(morphology,
                                         root,
                                         node_types=None):
    """
        Calculate the greatest number of branches encountered among all
        directed paths from the morphology's root to its leaves. A branch is
        defined as a root->leaf ordered path for which:
            1. the first node on the path is either
                a. a bifurcation (has > 1 children)
                b. the root node
            2. the last node on the path is either
                a. a bifurcation
                b. a leaf node (has 0 children)

        Parameters
        ----------
        morphology: the reconstruction whose max branch order will be
            calculated
        root: treat this node as root
        node_types: If not None, consider only root->leaf paths whose leaf
            nodes are among these types (see neuron_morphology constants)

        Returns
        -------
        The greatest branch count encountered among all considered root->leaf
        paths

    """

    root_id = morphology.node_id_cb(root)
    counter = {'branches_to_node': {root_id: 0},
               'max_branches': 0}

    def branch_visitor(node, counter, node_types):
        cur_branches = counter['branches_to_node'][node['id']]
        children = morphology.get_children(node)
        num_children = len(children)

        if num_children > 1:
            for child in children:
                counter['branches_to_node'][child['id']] = \
                    cur_branches + 1
        elif num_children == 1:
            if cur_branches == 0:
                # create a branch even if its just the root + one node
                counter['branches_to_node'][children[0]['id']] = 1
            else:
                counter['branches_to_node'][children[0]['id']] = cur_branches
        elif num_children == 0 and (node_types is None or node['type'] in node_types):
            if cur_branches > counter['max_branches']:
                counter['max_branches'] = cur_branches

    visitor = partial(branch_visitor,
                      counter=counter,
                      node_types=node_types)
    morphology.depth_first_traversal(visitor,
                                     start_id=root_id)


    return counter['max_branches']


@marked(Intrinsic)
def max_branch_order(
        data: Data,
        node_types: Optional[List] = None,
        ):
    """
        Calculate mean fragmentation

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

    """
    morphology = data.morphology
    roots = morphology.get_roots()
    max_branch_order = 0
    for root in roots:
        local_max = calculate_max_branch_order_from_root(
                morphology, root, node_types=node_types)
        if local_max > max_branch_order:
            max_branch_order = local_max

    return max_branch_order



from unittest import TestCase

import numpy as np

from neuron_morphology.snap_polygons.bounding_box import BoundingBox


class TestBoundingBox(TestCase):

    def setUp(self):
        self.box = BoundingBox(10, 20, 30, 40)

    def test_update(self):
        self.box.update(20, 15, 35, 40)

        assert np.allclose(
            self.box.coordinates,
            [10, 15, 35, 40]
        )

    def test_transform(self):
        new = self.box.transform( lambda v, h: (v *2, h) )
        assert np.allclose(
            new.coordinates,
            [20, 20, 60, 40]
        )

    def test_transform_inplace(self):
        self.box.transform( lambda v, h: (v *2, h) , inplace=True)
        assert np.allclose(
            self.box.coordinates,
            [20, 20, 60, 40]
        )

    def test_copy(self):
        new = self.box.copy()
        new.vorigin = 12
        self.assertEqual(self.box.vorigin, 20)

    def test_round(self):
        box = BoundingBox(10.5, 20, 30, 40.5)
        new = box.round(origin_via=np.floor, extent_via=np.ceil)
        assert np.allclose(
            new.coordinates,
            [10, 20, 30, 41]
        )


import unittest

import neuron_morphology.features.path as path
from neuron_morphology.morphology import Morphology
from neuron_morphology.constants import (
    SOMA, AXON, APICAL_DENDRITE, BASAL_DENDRITE)
from neuron_morphology.feature_extractor.feature_extractor import (
    FeatureExtractor
)
from neuron_morphology.feature_extractor.feature_specialization import (
    NEURITE_SPECIALIZATIONS)
from neuron_morphology.feature_extractor.marked_feature import specialize
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.morphology_builder import MorphologyBuilder

class PathTestCase(unittest.TestCase):
    def setUp(self):
        self.morphology = Morphology([
                {
                    "id": 0,
                    "parent_id": -1,
                    "type": SOMA,
                    "x": 0,
                    "y": 0,
                    "z": 100,
                    "radius": 1
                },
                {
                    "id": 1,
                    "parent_id": 0,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 101,
                    "radius": 1
                },
                {
                    "id": 2,
                    "parent_id": 0,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 102,
                    "radius": 1
                },
                { # bifurcates
                    "id": 3,
                    "parent_id": 1,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 110,
                    "radius": 1
                },                
                {
                    "id": 4,
                    "parent_id": 3,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 140,
                    "radius": 1 
                },     
                { # this node backtracks, causing the path distance to differ from the euclidean
                    "id": 11,
                    "parent_id": 4,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 130,
                    "radius": 1 
                },       
                { # bifurcates
                    "id": 5,
                    "parent_id": 3,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 130,
                    "radius": 1 
                },                
                {
                    "id": 6,
                    "parent_id": 5,
                    "type": AXON,
                    "x": 0,
                    "y": 0,
                    "z": 135,
                    "radius": 1 
                },       
                {
                    "id": 7,
                    "parent_id": 5,
                    "type": AXON,
                    "x": 30,
                    "y": 0,
                    "z": 103,
                    "radius": 1 
                },     
                { # bifurcates
                    "id": 8,
                    "parent_id": 2,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 125,
                    "radius": 1 
                },
                {
                    "id": 9,
                    "parent_id": 8,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 126,
                    "radius": 1 
                },
                {
                    "id": 10,
                    "parent_id": 8,
                    "type": APICAL_DENDRITE,
                    "x": 0,
                    "y": 0,
                    "z": 127,
                    "radius": 1 
                },
            ],
            node_id_cb=lambda node: node["id"],
            parent_id_cb=lambda node: node["parent_id"],
        )


class TestMeanContraction(PathTestCase):

    def test_generic(self):
        self.assertAlmostEqual(
            path.mean_contraction(self.morphology),
            0.86049192013635
        )

    def test_specific(self):
        self.assertAlmostEqual(
            path.mean_contraction(
                self.morphology, node_types=[APICAL_DENDRITE]),
            1.0 # only axons backtrack
        )

    def test_specialized(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(2, 1, 0)
                                .basal_dendrite(1, 1, 0).up()
                                .basal_dendrite(3, 1, 0).up(2)
                            .axon(2, -1, 0)
                                .axon(2, -2, 0)
                                    .axon(1, -2, 0).up()
                                    .axon(3, -2, 0).up(5)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 2, 0).up()
                            .apical_dendrite(1, 2, 0)
                .build()
        )

        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(path.mean_contraction, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["axon.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["dendrite.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.mean_contraction"], 1)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.mean_contraction"], 1)


class TestEarlyBranchPath(PathTestCase):

    def test_generic(self):

        self.assertAlmostEqual(
            path.early_branch_path(self.morphology),
            40 / 70.36087214122114 
            # longest short path distance is 40 on the path starting 3 -> 4
            # 70. ...  is the max_path_distance of this tree
        )


class TestMaxPathDistance(PathTestCase):

    def test_generic(self):
        self.assertAlmostEqual(
            path.max_path_distance(self.morphology),
            70.36087214122114
        )

    def test_specialized(self):
        morphology = (
            MorphologyBuilder()
                .root(0, 0, 0)
                    .basal_dendrite(1, 0, 0)
                        .axon(1, -1, 0).up()
                        .basal_dendrite(2, 0, 0)
                            .basal_dendrite(3, 0, 0).up()
                            .basal_dendrite(2, 1, 0).up(3)
                    .apical_dendrite(0, 1, 0)
                        .apical_dendrite(0, 2, 0)
                            .apical_dendrite(-1, 2, 0).up()
                            .apical_dendrite(1, 2, 0)
                                .apical_dendrite(1, 3, 0)
                .build()
        )


        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(path.max_path_distance, NEURITE_SPECIALIZATIONS)])
        feature_extraction_run = fe.extract(cell_data)

        self.assertEqual(feature_extraction_run.results["all_neurites.max_path_distance"], 4)
        self.assertEqual(feature_extraction_run.results["axon.max_path_distance"], 2)
        self.assertEqual(feature_extraction_run.results["dendrite.max_path_distance"], 4)
        self.assertEqual(feature_extraction_run.results["basal_dendrite.max_path_distance"], 3)
        self.assertEqual(feature_extraction_run.results["apical_dendrite.max_path_distance"], 4)



import unittest

from neuron_morphology.constants import *
from neuron_morphology.morphology import Morphology
from neuron_morphology.feature_extractor.feature_extractor import FeatureExtractor # for individual cell features
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.marked_feature import (
    specialize
)
from neuron_morphology.feature_extractor.feature_specialization import (
    BasalDendriteSpec, AxonSpec
)
from neuron_morphology.features.soma import (
    calculate_stem_exit_and_distance
)
from tests.objects import test_node


class TestStemExit(unittest.TestCase):

    def test_stem_exit_with_type(self):

        nodes = [
            test_node(id=1, type=SOMA, x=0, y=0, z=0, radius=10, parent_node_id=-1),
            test_node(id=2, type=BASAL_DENDRITE, x=10, y=0, z=0, radius=2, parent_node_id=1),
            test_node(id=3, type=BASAL_DENDRITE, x=20, y=0, z=0, radius=2, parent_node_id=2),
            test_node(id=4, type=AXON, x=30, y=0, z=0, radius=0.5, parent_node_id=2),
            test_node(id=5, type=AXON, x=40, y=0, z=0, radius=0.5, parent_node_id=4),
            test_node(id=6, type=BASAL_DENDRITE, x=0, y=-10, z=0, radius=2, parent_node_id=1),
        ]

        for node in nodes:
            node['parent'] = int(node['parent'])
            node['id'] = int(node['id'])
            node['type'] = int(node['type'])

        morphology = Morphology(nodes, node_id_cb=lambda node: node['id'], parent_id_cb=lambda node: node['parent'])


        cell_data = Data(morphology=morphology)
        fe = FeatureExtractor()
        fe.register_features([specialize(calculate_stem_exit_and_distance, [BasalDendriteSpec, AxonSpec])])
        feature_extraction_run = fe.extract(cell_data)

        axon_results = feature_extraction_run.results["axon.calculate_stem_exit_and_distance"]
        basal_results = feature_extraction_run.results["basal_dendrite.calculate_stem_exit_and_distance"]
        self.assertEqual(axon_results[0][0], 0.5)
        self.assertEqual(axon_results[0][1], 30.0)
        self.assertEqual(basal_results[0][0], 0.5)
        self.assertEqual(basal_results[0][1], 0)
        self.assertEqual(basal_results[1][0], 1)
        self.assertEqual(basal_results[1][1], 0)



from typing import Optional, List, Union, Dict, Any
from statistics import mean
from collections import defaultdict
from functools import partial

from neuron_morphology.morphology import Morphology
from neuron_morphology.constants import SOMA
from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.mark import (
    Geometric, RequiresRadii, RequiresRoot)
from neuron_morphology.feature_extractor.data import (
    MorphologyLike, get_morphology)

@marked(Geometric)
def total_length(
    data: MorphologyLike, 
    node_types: Optional[List[int]] = None
) -> float:
    """ Calculate the total length across all compartments in a reconstruction

    Parameters
    ----------
    data : the input reconstruction
    node_types : if provided, restrict the calculation to compartments 
        involving these types

    Returns
    -------
    The sum of segment lengths across all segments in the reconstruction

    Notes
    -----
    Excludes compartments where the parent is:
        1. the soma
        2. a root of the reconstruction
    The logic here is that the soma root is likely to substantially overlap any 
    of its compartments, while non-root soma nodes will be closer to the soma 
    surface.

    """

    morphology = get_morphology(data)

    nodes = morphology.get_node_by_types(node_types)
    compartment_list = morphology.get_compartments(nodes, node_types)

    total = 0.0
    for compartment in compartment_list:
        first_node_in_compartment = compartment[0]
        if first_node_in_compartment['type'] is SOMA \
            and not morphology.parent_of(first_node_in_compartment):
            continue
        total += morphology.get_compartment_length(compartment)

    return total


@marked(RequiresRadii)
@marked(Geometric)
def total_surface_area(
    data: MorphologyLike, 
    node_types: Optional[List[int]] = None
) -> float:
    """ Calculates the sum of lateral surface areas across all comparments 
    (linked pairs of nodes) in a reconstruction. This approximates the total 
    surface area of the reconstruction. See 
    Morphology.get_compartment_surface_area for details.

    Parameters
    ----------
    data : The reconstruction whose surface area will be computed
    node_types : restrict the calculation to compartments involving these node 
        types

    Returns
    -------
    The sum of compartment lateral surface areas across this reconstruction

    """

    morphology: Morphology = get_morphology(data)
    nodes = morphology.get_node_by_types(node_types)
    compartments = morphology.get_compartments(nodes, node_types)

    return sum(map(morphology.get_compartment_surface_area, compartments))


@marked(RequiresRadii)
@marked(Geometric)
def total_volume(
    data: MorphologyLike, 
    node_types: Optional[List[int]] = None
) -> float:
    """ Calculates the sum of volumes across all comparments (linked pairs of 
    nodes) in a reconstruction. This approximates the total volume of the 
    reconstruction. See Morphology.get_compartment_volume for details.

    Parameters
    ----------
    data : The reconstruction whose volume will be computed
    node_types : restrict the calculation to compartments involving these node 
        types

    Returns
    -------
    The sum of compartment volumes across this reconstruction

    """
    
    morphology = get_morphology(data)
    nodes = morphology.get_node_by_types(node_types)
    compartments = morphology.get_compartments(nodes, node_types)

    return sum(map(morphology.get_compartment_volume, compartments))


@marked(RequiresRadii)
def mean_diameter(
    data: MorphologyLike, 
    node_types: Optional[List[int]] = None
) -> float:
    """ Calculates the mean diameter of all nodes
    
    Parameters
    ----------
    morphology : The reconstruction whose mean diameter
    node_types : restrict the calculation to compartments involving these node 
        types

    Returns
    -------
    The average diameter across selected nodes

    """

    morphology = get_morphology(data)

    return 2 * mean(
        node["radius"] for node in morphology.get_node_by_types(node_types)
    )



def parent_daughter_ratio_visitor(
    node: Dict[str, Any], 
    morphology: Morphology, 
    counters: Dict[str, Union[int, float]],
    node_types: Optional[List[int]] = None
):
    """ Calculates for a single node the ratio of the node's parent's radius to 
    the node's radius. Stores these values in a provided dictionary.

    Parameters
    ----------
    node : The node under consideration
    morphology : The reconstruction to which this node belongs
    counters : a dictionary used for storing running ratio totals and counts.
    node_types : skip nodes not of one of these types

    Notes
    -----
    see mean_parent_daughter_ratio for usage

    """

    parent = morphology.parent_of(node)
    
    if parent is None:
        return

    if node_types is not None:
        if node["type"] not in node_types or parent["type"] not in node_types:
            return

    counters["ratio_sum"] += parent["radius"] / node["radius"]
    counters["ratio_count"] += 1
    

@marked(RequiresRadii)
def mean_parent_daughter_ratio(
    data: MorphologyLike,
    node_types: Optional[List[int]] = None
) -> float:
    """ Calculate the average ratio of parent radii to child radii across a 
    reconstruction.

    Parameters
    ----------
    data : The reconstruction whose mean parent daugther ratio will be computed
    node_types : restrict the calculation to compartments involving these node 
        types

    Notes
    -----
    Note that this function differs from the L-measure parent daughter ratio, 
    which calculates the ratio of the child node size to the parent. Note also 
    that both the parent and child must be in node_types in order for a
    compartment to be included in the calculation

    """

    morphology = get_morphology(data)
    roots = morphology.get_roots()
    
    counters: Dict[str, int] = defaultdict(lambda *a, **k: 0)
    visitor = partial(
        parent_daughter_ratio_visitor, 
        morphology=morphology, 
        counters=counters,
        node_types=node_types
    )
    
    for root in roots:
        morphology.breadth_first_traversal(
            visitor, 
            start_id=morphology.node_id_cb(root)
        )

    return counters["ratio_sum"] / counters["ratio_count"]


@marked(RequiresRoot)
@marked(Geometric)
def max_euclidean_distance(
    data: MorphologyLike,
    node_types: Optional[List[int]] = None
) -> float:
    """Calculate the furthest distance, in 3-space, of a compartment's end from 
    the soma. This is equivalent to the distance to the furthest SWC node.

    Parameters
    ----------
    data: The reconstruction whose max euclidean distance will be 
        calculated
    node_types: restrict consideration to these types

    Returns
    -------
    The distance between the soma and the farthest-from-soma node in this 
    morphology.

    """

    morphology = get_morphology(data)
    soma = morphology.get_root()

    return max(
        morphology.euclidean_distance(soma, node)
        for node in morphology.get_node_by_types(node_types)
    )



import abc

from neuron_morphology.morphology import Morphology


class TransformBase(abc.ABC):
    """
        Abstract base class for implementing swc transforms.
        Each child class should implement these methods.
    """
    @abc.abstractmethod
    def transform_morphology(self) -> Morphology:
        """
            Apply this transform to all nodes in a morphology.

            Returns
            -------
            A Morphology

        """
        raise NotImplementedError()

    @abc.abstractmethod
    def transform(self):
        """
            Apply this transform to (3,) point or (3,n) array-like of points.

            Returns
            -------
            numpy.ndarray with same shape as input

        """
        raise NotImplementedError()



import unittest
import tempfile
import shutil
import os
import subprocess as sp
import json
import os

import pandas as pd
import numpy as np
import h5py
import pytest

import neuron_morphology.feature_extractor.__main__ as main
from neuron_morphology.swc_io import write_swc
from neuron_morphology.constants import (
    SOMA, APICAL_DENDRITE, BASAL_DENDRITE, AXON
)
from neuron_morphology.morphology_builder import MorphologyBuilder
from neuron_morphology.features.layer.layered_point_depths import \
    LayeredPointDepths


TIMEOUT = int(os.getenv("TIMEOUT", "60"))


def nodes():
    return (
        MorphologyBuilder()
            .root()
                .axon()
                    .axon()
                        .axon().up(3)
                .apical_dendrite()
                    .apical_dendrite()
                        .apical_dendrite().up()
                        .apical_dendrite()
            .nodes
    )


class TestRun(unittest.TestCase):

    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

        _nodes = nodes()
        self.first_path = os.path.join(self.tmpdir, "first.swc")
        write_swc(pd.DataFrame(nodes()), self.first_path)
        self.second_path = os.path.join(self.tmpdir, "second.swc")
        write_swc(pd.DataFrame(nodes()), self.second_path)

        self.lpd_path = os.path.join(self.tmpdir, "layered_point_depths.csv")
        LayeredPointDepths(
            ids=np.arange(8)[::-1],
            layer_name=["2", "2", "wm", "wm", "2", "1", "1", "1"],
            depth=[200, 230, 260, 290, 60, 40, 30, 20],
            local_layer_pia_side_depth=[50, 50, 250, 250, 50, 0, 0, 0],
            local_layer_wm_side_depth=[250, 250, np.nan, np.nan, 250, 50, 50, 50],
            point_type=[node["type"] for node in _nodes]
        ).to_csv(self.lpd_path)

        self.heavy_output_path = os.path.join(self.tmpdir, "heavy.h5")
        input_json_data = {
            "heavy_output_path": self.heavy_output_path,
            "reconstructions": [
                {
                    "swc_path": self.first_path,
                    "identifier": "first",
                    "layered_point_depths_path": self.lpd_path
                },
                {
                    "swc_path": self.second_path,
                    "identifier": "second"
                }
            ],
            "global_parameters": {
                "reference_layer_depths": {
                    "names": ["1", "2", "wm"],
                    "boundaries": [0, 100, 200, 300]
                }
            }
        }
        self.input_json_path = os.path.join(self.tmpdir, "input.json")
        self.output_json_path = os.path.join(self.tmpdir, "output.json")
        with open(self.input_json_path, "w") as input_file:
            json.dump(input_json_data, input_file)

        sp.check_call([
            "python", "-m", "neuron_morphology.feature_extractor", 
            "--input_json", self.input_json_path,
            "--output_json", self.output_json_path
            ],
            timeout=TIMEOUT)

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    # we are invoking our tests in ci using pytest, resulting in this hideous 
    # mixing of frameworks - unittest can also skipIf but the directive is not 
    # respected by pytest collector
    @pytest.mark.skipif(TIMEOUT == 0, reason="potentially long-running test")
    def test_run(self):

        with open(self.output_json_path, "r") as output_file:
            obtained = json.load(output_file)

        first_results = obtained["results"]["first"]["results"]
        second_results = obtained["results"]["second"]["results"]

        self.assertEqual(first_results["axon.num_tips"], 1)
        self.assertEqual(
            first_results
                ["axon.apical_dendrite.earth_movers_distance.2"]
                ["result"],
            17.0
        )
        self.assertEqual(
            first_results
                ["axon.apical_dendrite.earth_movers_distance.2"]
                ["interpretation"],
            "BothPresent"
        )

        self.assertNotIn(
            "axon.apical_dendrite.earth_movers_distance.2", second_results)

    @pytest.mark.skipif(TIMEOUT == 0, reason="potentially long-running test")
    def test_run_heavy_out(self):
        expected_counts = np.zeros(20)
        expected_counts[18] = 1

        with h5py.File(self.heavy_output_path, "r") as hf:
            counts = hf["first/axon.normalized_depth_histogram.2/counts"][:]
            assert np.allclose(counts, expected_counts)



import math
import numpy as np

from functools import partial
from typing import Optional, List, Dict
from neuron_morphology.feature_extractor.marked_feature import (
    MarkedFeature, marked
)
from neuron_morphology.feature_extractor.mark import (
    RequiresRoot,
    RequiresSoma,
    RequiresRelativeSomaDepth,
    RequiresApical,
    RequiresBasal,
    RequiresAxon,
    RequiresDendrite,
    Geometric
)
from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.constants import (
    SOMA, AXON, BASAL_DENDRITE, APICAL_DENDRITE
)
from neuron_morphology.morphology import Morphology


__all__ = [
    "calculate_soma_surface",
    "calculate_relative_soma_depth",
    "calculate_soma_features",
    "calculate_stem_exit_and_distance"
]


@marked(Geometric)
@marked(RequiresRoot)
def calculate_soma_surface(data: Data) -> float:

    """
        Approximates the surface area of the soma. Morphologies with only
        a single soma node are supported.

        Parameters
        ----------
        data: Data Object containing a morphology

        Returns
        -------

        Scalar value

    """

    soma = data.morphology.get_soma()
    return 4.0 * math.pi * soma['radius'] * soma['radius']


@marked(RequiresRoot)
@marked(RequiresRelativeSomaDepth)
def calculate_relative_soma_depth(data: Data) -> float:
    """
        Calculate the soma depth relative to pia/wm

        Parameters
        ----------
        data: Data Object containing a morphology

        Returns
        -------

        Scalar value

    """

    return data.relative_soma_depth


@marked(Geometric)
@marked(RequiresRoot)
@marked(RequiresRelativeSomaDepth)
def calculate_soma_features(data: Data):
    """
        Calculate the soma features

        Parameters
        ----------
        data: Data Object containing a morphology

        Returns
        -------

        soma_features

    """

    features = {}
    features["soma_surface"] = calculate_soma_surface(data.morphology)
    features["relative_soma_depth"] = calculate_relative_soma_depth(data)

    return features


@marked(Geometric)
@marked(RequiresSoma)
@marked(RequiresRoot)
def calculate_stem_exit_and_distance(data: Data, node_types: Optional[List[int]], z_scale=3.0):
    """
        Returns the relative radial position (stem_exit) on the soma where the
        tree holding the tree connects to the soma. 0 is on the bottom,
        1 on the top, and 0.5 out a side.
        Also returns the distance (stem_distance) between the tree root and the
        soma surface.

        Parameters
        ----------

        data: Data Object containing a morphology

        soma: dict
        soma node

        node_types: list (AXON, BASAL_DENDRITE, APICAL_DENDRITE)
        Type to restrict search to

        Returns
        -------

        (float, float):
        First value is relative position (height, on [0,1]) of a
        tree on soma. Second value is distance of the root from soma

    """

    # find axon node, get its tree ID, fetch that tree, and see where
    #   it connects to the soma radially
    nodes = data.morphology.get_node_by_types(node_types)

    node_ids = [data.morphology.node_id_cb(n) for n in nodes]
    tree_root = None
    stem_distance = 0
    stem_exit = 0

    soma = data.morphology.get_soma()

    # find all root nodes (meaning a node whose parent is not in the set of nodes under consideration)
    root_nodes = []
    for node in nodes:
        if data.morphology.parent_id_cb(node) not in node_ids:
            root_nodes.append(node)

    # set the root node closest to soma as the tree_root
    stem_exit_info = []
    for node in root_nodes:
        if data.morphology.parent_id_cb(node) == data.morphology.node_id_cb(soma):
            stem_distance = 0
        else:
            stem_distance = data.morphology.euclidean_distance(node, soma)

        # make point soma-radius north of soma root
        # do acos(dot product) to get angle of tree root from vertical
        # adjust so 0 is theta=pi and 1 is theta=0
        vert = np.zeros(3)
        vert[1] = 1.0
        root = np.zeros(3)
        root[0] = node['x'] - soma['x']
        root[1] = node['y'] - soma['y']
        root[2] = (node['z'] - soma['z']) * z_scale
        stem_exit = np.arccos(np.clip(np.dot(vert/np.linalg.norm(vert), root/np.linalg.norm(root)), -1.0, 1.0)) / math.pi

        stem_exit_info.append((stem_exit, stem_distance))

    return stem_exit_info


@marked(RequiresSoma)
@marked(RequiresRoot)
def calculate_number_of_stems(data: Data, node_types: Optional[List[int]]):

    """
        Calculate the number of soma stems.
        This is defined as the total number of non-soma child nodes on soma nodes.

        Parameters
        ----------
        data: Data Object containing a morphology

        node_types: list (AXON, BASAL_DENDRITE, APICAL_DENDRITE)
        Type to restrict search to

        Returns
        -------

        Scalar value

    """

    soma = data.morphology.get_soma()
    if node_types:
        return len(data.morphology.get_children_of_node_by_types(soma, node_types))
    else:
        return len(data.morphology.children_of(soma))

@marked(Geometric)
@marked(RequiresRoot)
def soma_percentile(data: Data,
                    node_types: Optional[List[int]],
                    symmetrize_xz: bool = True):
    """
        Calculates the percent of of nodes that are below the soma.
        If symmetrize_xz is true, then p

        Parameters
        ----------

        data: Data Object containing a morphology

        node_types: a list of node types (see neuron_morphology constants)

        symmetrize_xz: bool indicating that x and z percentages should always
                       fall between 0 and 0.5 to prevent handedness
                       (e.g if 0.75 nodes are below the soma in the
                        x direction, then return 1-0.75=0.25 instead)


        Returns
        -------

        percentiles: array of x, y, and z percentiles

    """
    soma_node = data.morphology.get_root()
    soma_coord = np.asarray([soma_node['x'], soma_node['y'], soma_node['z']])

    nodes = data.morphology.get_node_by_types(node_types=node_types)
    coords = np.asarray([[node['x'], node['y'], node['z']] for node in nodes])

    num_less_than = coords < soma_coord
    percentile = num_less_than.sum(axis=0) / num_less_than.shape[0]

    if symmetrize_xz:
        if percentile[0] > 0.5:
            percentile[0] = 1.0 - percentile[0]
        if percentile[2] > 0.5:
            percentile[2] = 1.0 - percentile[2]

    return percentile



import unittest

from neuron_morphology.transforms.transform_base import TransformBase


class TestTransformBase(unittest.TestCase):

    def test_cannot_instantiate(self):
        with self.assertRaises(TypeError):
            TransformBase()

    def test_missing_transform_method(self):

        class ConcreteTransform(TransformBase):
            def transform_morphology(self):
                return "implements transform_morphology but not transform"

        with self.assertRaises(TypeError):
            ConcreteTransform()

    def test_missing_transform_morphology_method(self):

        class ConcreteTransform(TransformBase):
            def transform(self):
                return "implements transform but not transform_morphology"

        with self.assertRaises(TypeError):
            ConcreteTransform()

    def test_concrete(self):

        class ConcreteTransform(TransformBase):
            def transform_morphology(self):
                return "implements transform_morphology"

            def transform(self):
                return "implements transform"

        ConcreteTransform()
        assert True



import unittest
import sys

from shapely.geometry import LineString, Point

import neuron_morphology.snap_polygons.cortex_surfaces as cs


class TestStandalone(unittest.TestCase):
    """Tests for functions operating on shapely objects
    """

    def test_remove_duplicates(self):
        cases = [
            [
                [Point(0, 1), Point(0, 1), Point(1, 1), Point(0, 1)],
                [Point(0, 1), Point(1, 1), Point(0, 1)]
            ],
            [
                [Point(0, 1), Point(0, 0), Point(0, 0), Point(0, 0)],
                [Point(0, 1), Point(0, 0)]
            ],
            [
                [Point(0, 1), Point(0.0, 1.0), Point(1, 1)],
                [Point(0, 1), Point(1, 1)]
            ],
            [
                [Point(0, 1), Point(sys.float_info.epsilon, 1), Point(1, 1)],
                [Point(0, 1), Point(sys.float_info.epsilon, 1), Point(1, 1)]
            ]
        ]

        for inpt, expected in cases:
            with self.subTest():
                obtained = cs.remove_duplicates(inpt)
                self.assertEqual(LineString(obtained), LineString(expected))

    def test_find_transition(self):
        condition = lambda pt: Point(0, 0).distance(pt) <= 1
        cases = [
            [Point(0, -2), Point(0, 0), 0, Point(0, -1)],
            [Point(0, -3), Point(0, 0), 0, Point(0, 0)],
            [Point(0, -4), Point(0, 0), 2, Point(0, -1)],
            [Point(0, -4), Point(0, -1), 10, Point(0, -1)],
            [Point(0, -1.5), Point(0, 0.5), 1, Point(0, -1)]
        ]

        for unmet, met, iterations, expected in cases:
            with self.subTest():
                obtained = cs.find_transition(
                    unmet, met, condition, iterations)
                print(obtained.xy, expected.xy)
                self.assertEqual(obtained, expected)


class TestCompound(unittest.TestCase):
    """Tests for functions operating on other cortex_surfaces functions
    """

    def test_first_met(self):
        condition = lambda pt: Point(0, 0).distance(pt) <= 1
        cases = [
            [
                [(-4, 0), (-3, 0), (-2, 0), (-1, 0), (0, 0)],
                2,
                3,
                Point(-1, 0)
            ],
            [
                [Point(0, 0), Point(1, 0), Point(2, 0)],
                30,
                0,
                Point(0, 0)
            ]
        ]

        for coords, iterations, idx_exp, coord_exp in cases:
            with self.subTest():
                idx_obt, coord_obt = cs.first_met(
                    coords, condition, iterations)
                print(idx_obt, idx_exp, coord_obt, coord_exp)
                self.assertEqual(idx_obt, idx_exp)
                self.assertEqual(coord_obt, coord_exp)

    def test_trim_coords(self):
        condition = lambda pt: Point(0, 0).distance(pt) <= 1
        cases = [
            [
                [Point(-2, 0), Point(-1, 0), Point(0, 0), Point(1, 0), Point(2, 0)],
                10,
                [Point(-1, 0), Point(0, 0), Point(1, 0)]
            ],
            [
                [Point(-2.5, 0), Point(-1.5, 0), Point(0.5, 0), Point(1.5, 0), Point(2.5, 0)],
                10,
                [Point(-1.0, 0), Point(0.5, 0), Point(1.0, 0)]
            ]
        ]

        for coords, iterations, expected in cases:
            with self.subTest():
                obtained = cs.trim_coords(coords, condition, iterations)
                ls_obt = LineString(obtained)
                ls_exp = LineString(expected)
                print(ls_obt, ls_exp)
                self.assertEqual(ls_obt, ls_exp)

    def trim_to_close(self):
        cases = [
            [
                [Point(-2, 0), Point(-1, 0), Point(0, 0), Point(1, 0), Point(2, 0)],
                1.0,
                10,
                [Point(-1, 0), Point(0, 0), Point(1, 0)]
            ],
            [
                [Point(-2.5, 0), Point(-1.5, 0), Point(0.5, 0), Point(1.5, 0), Point(2.5, 0)],
                1.0,
                10,
                [Point(-1.0, 0), Point(0.5, 0), Point(1.0, 0)]
            ],
            [
                [Point(-2.5, 0), Point(-1.5, 0), Point(0.5, 0), Point(1.5, 0), Point(2.5, 0)],
                2.0,
                10,
                [Point(-2.0, 0), Point(-1.5, 0), Point(0.5, 0), Point(1.5, 0), Point(2.0, 0)]
            ]
        ]

        for coords, threshold, iterations, expected in cases:
            with self.subTest():
                obtained = cs.trim_to_close(Point(0, 0), threshold, LineString(coords), iterations)
                ls_obt = LineString(obtained)
                ls_exp = LineString(expected)
                print(ls_obt, ls_exp)
                self.assertEqual(ls_obt, ls_exp)


from neuron_morphology.validation.result import NodeValidationError as ve
from neuron_morphology.constants import *

valid_types = {SOMA, AXON, BASAL_DENDRITE, APICAL_DENDRITE}


def validate_count_node_parent(morphology, node_type, parent_type, expected_count):

    """ This function validates the number of nodes that have a specific type of parent """

    matched_node_numbers = []
    result = []

    nodes_by_type = morphology.get_node_by_types([node_type])
    for node in nodes_by_type:
        if morphology.parent_of(node) and morphology.parent_of(node)['type'] is parent_type:
            matched_node_numbers.append(node['id'])

    if len(matched_node_numbers) > expected_count:
        for node_number in matched_node_numbers:
            result.append(ve("Nodes of type %s can only have %s parent of type %s" % (node_type, expected_count
                                                                                      , parent_type), node_number
                                                                                                    , "Warning"))

    return result


def validate_number_of_soma_nodes(morphology):
    """ This function validates the number of type 1 nodes """

    result = []

    soma = morphology.get_node_by_types([SOMA])

    if len(soma) != 1:
        for node in soma:
            result.append(ve("The morphology needs to have one soma node", node['id'], "Error"))

    return result


def validate_expected_types(morphology):

    """ This function validates the expected types of the nodes """

    result = []

    nodes_with_invalid_type = morphology.filter_nodes(lambda node: node['type'] not in valid_types)
    for node in nodes_with_invalid_type:
        result.append(ve("Node type needs to be one of these values: %s" % valid_types, node['id'], "Warning"))

    return result


def valid_dendrite_parent(morphology, node, valid_parent_type):

    parent = morphology.parent_of(node)
    return parent and parent['type'] in valid_parent_type


def validate_node_parent(morphology):

    """ This function validates the type of parent node for a specific type of child node """

    result = []
    valid_soma_parents = {None}
    valid_axon_parents = {SOMA, AXON, BASAL_DENDRITE, None}
    valid_basal_dendrite_parents = {SOMA, BASAL_DENDRITE}
    valid_apical_dendrite_parents = {SOMA, APICAL_DENDRITE}

    soma_node_with_invalid_parents = morphology.filter_nodes(lambda node: node['type'] == SOMA and
                                                                          morphology.parent_of(node))
    if soma_node_with_invalid_parents:
        result.append(ve("Type 1 can only have a parent of the following types: %s" % valid_soma_parents,
                         soma_node_with_invalid_parents[0]['id'], "Error"))

    axon_nodes_with_invalid_parents = morphology.filter_nodes(lambda node: node['type'] == AXON and
                                                                           morphology.parent_of(node) and
                                                                           morphology.parent_of(node)['type']
                                                                           not in valid_axon_parents)

    for node in axon_nodes_with_invalid_parents:
        result.append(ve("Type 2 can only have a parent of the following types: %s" % valid_axon_parents, node['id'],
                         "Error"))

    basal_dendrite_nodes = morphology.get_node_by_types([BASAL_DENDRITE])

    for node in basal_dendrite_nodes:
        if not valid_dendrite_parent(morphology, node, valid_basal_dendrite_parents):
            result.append(ve("Type 3 can only have a parent of the following types: %s" % valid_basal_dendrite_parents,
                             node['id'], "Error"))

    apical_dendrite_nodes = morphology.get_node_by_types([APICAL_DENDRITE])
    for node in apical_dendrite_nodes:
        if not valid_dendrite_parent(morphology, node, valid_apical_dendrite_parents):
            result.append(ve("Type 4 can only have a parent of the following types: %s" % valid_apical_dendrite_parents,
                             node['id'], "Error"))

    return result


def validate_immediate_children_of_soma_cannot_branch(morphology):

    """ This function validates that immediate children of soma cannot branch """

    result = []

    soma = morphology.get_root()
    if soma:
        soma_children = morphology.children_of(soma)
        for node in soma_children:
            if len(morphology.children_of(node)) > 1:
                result.append(ve("Immediate children of soma cannot branch", node['id'], "Error"))

    return result


def validate_multiple_axon_initiation_points(morphology):

    """ This function validates that the parent of axon (either type 1 or 3) only happens once """

    nodes = morphology.nodes()
    expected_count = 1
    matched_node_numbers = []
    result = []
    for node in nodes:
        if node['type'] == AXON:
            parent = morphology.parent_of(node)
            if parent and parent['type'] in [BASAL_DENDRITE, SOMA]:
                matched_node_numbers.append(node['id'])

    if len(matched_node_numbers) > expected_count:
        for node_number in matched_node_numbers:
            result.append(ve("Axon can only have one parent of type basal dendrite or soma", node_number, "Error"))

    return result


def validate(morphology):

    result = []

    result += validate_expected_types(morphology)

    result += validate_node_parent(morphology)

    result += validate_immediate_children_of_soma_cannot_branch(morphology)

    """ nodes that are type 4 can only have one parent parent of type 1 """
    result += validate_count_node_parent(morphology, APICAL_DENDRITE, SOMA, 1)

    """ There can only be one node of type 1 """
    result += validate_number_of_soma_nodes(morphology)

    result += validate_multiple_axon_initiation_points(morphology)

    return result



import unittest

import numpy as np

from neuron_morphology.features.layer.reference_layer_depths import \
    ReferenceLayerDepths


class TestReferenceLayerDepths(unittest.TestCase):

    def test_thickness(self):
        depths = ReferenceLayerDepths(1, 2)
        self.assertEqual(depths.thickness, 1)

    def test_sequential(self):
        depths = ReferenceLayerDepths.sequential(["a", "b", "wm"], [1, 2, 3, 4])
        
        assert np.allclose([depth.pia_side for depth in depths.values()], 
            [1, 2, 3])
        assert np.allclose([depth.wm_side for depth in depths.values()], 
            [2, 3, 4])
        assert np.allclose([depth.scale for depth in depths.values()], 
            [True, True, False])
        self.assertEqual(set(depths.keys()), {"a", "b", "wm"})

    def test_sequential_mismatched(self):
        with self.assertRaises(AssertionError):
            ReferenceLayerDepths.sequential(["a", "b", "wm"], [1, 2, 3, 4, 5])

    def test_sequential_all_scale(self):
        depths = ReferenceLayerDepths.sequential(["a", "b", "wm"], 
            [1, 2, 3, 4], last_is_scale=True)
        assert np.allclose([depth.scale for depth in depths.values()], 
            [True, True, True])


import sys

from argschema import ArgSchemaParser

from neuron_morphology.swc_io import morphology_from_swc, morphology_to_swc
from neuron_morphology.transforms.affine_transformer._schemas import (
    ApplyAffineSchema, OutputParameters)
from neuron_morphology.transforms.affine_transform import AffineTransform


def main():
    mod = ArgSchemaParser(schema_type=ApplyAffineSchema,
                          output_schema_type=OutputParameters)
    args = mod.args

    if 'affine_dict' in args:
        affine_transform = AffineTransform.from_dict(args['affine_dict'])
    elif 'affine_list' in args:
        affine_transform = AffineTransform.from_list(args['affine_list'])
    else:
        raise ValueError('must provide either an affine_dict or affine_list')

    morph_in = morphology_from_swc(args['input_swc'])

    morph_out = affine_transform.transform_morphology(morph_in)

    morphology_to_swc(morph_out, args['output_swc'])

    output = {
        'inputs': args,
        'transformed_swc': args['output_swc'],
    }

    mod.output(output)


if __name__ == '__main__':
    main()



# Copyright 2015-2017 Allen Institute for Brain Science
# This file is part of Allen SDK.
#
# Allen SDK is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, version 3 of the License.
#
# Allen SDK is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Allen SDK.  If not, see <http://www.gnu.org/licenses/>.
# Author: Nika Keller


class NodeValidationError(object):

    def __init__(self, message, node_ids, level):
        self._message = message
        if type(node_ids) is list:
            self._node_ids = node_ids
        else:
            self._node_ids = [node_ids]
        self._level = level

    @property
    def message(self):
        return self._message

    @property
    def node_ids(self):
        return self._node_ids

    @property
    def level(self):
        return self._level

    def __repr__(self):
        return "Message: %s, Node ID: %s, Level: %s" % (self._message, self._node_ids, self.level)


class MarkerValidationError(object):

    def __init__(self, message, marker, level):
        self._message = message
        if type(marker) is list:
            self._marker = marker
        else:
            self._marker = [marker]
        self._level = level

    @property
    def message(self):
        return self._message

    @property
    def marker(self):
        return self._marker

    @property
    def level(self):
        return self._level

    def __repr__(self):
        return "Message: %s, Marker: %s, Level: %s" % (self._message, self._marker, self._level)


class InvalidMorphology(ValueError):

    def __init__(self, validation_errors):
        super(InvalidMorphology, self).__init__("Morphology appears to be inconsistent")
        self._validation_errors = validation_errors

    @property
    def validation_errors(self):
        return self._validation_errors

    def __str__(self):
        return "%s, Errors: %s" % (ValueError.__str__(self), self._validation_errors.__str__())


class InvalidMarkerFile(ValueError):

    def __init__(self, validation_errors):
        super(InvalidMarkerFile, self).__init__("Marker file is not valid")
        self._validation_errors = validation_errors

    @property
    def validation_errors(self):
        return self._validation_errors

    def __str__(self):
        return "%s, Errors: %s" % (ValueError.__str__(self), self._validation_errors.__str__())



import os
import sys
import subprocess as sp
import unittest
import shutil
import tempfile
import xarray as xr
import numpy as np

try:
    import dolfinx

    has_dolfinx = True
except ImportError:
    has_dolfinx = False


@unittest.skipIf(
    not has_dolfinx,
    "streamline calculation requires conda-installed fenics-dolfinx",
)
class TestRunPiaWmStreamlines(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp()
        self.input = {
            "pia_path_str": "120,100,0,100",
            "wm_path_str": "20,0,100,0",
            "soma_path_str": "50,50",
            "output_dir": self.test_dir,
            "output_json": os.path.join(self.test_dir, "output.json"),
        }
        self.output = {
            "input": self.input,
            "depth_field_file": os.path.join(self.test_dir, "depth_field.nc"),
            "gradient_field_file": os.path.join(self.test_dir, "gradient_field.nc"),
        }

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_end_to_end(self):
        cmd = [
            "python",
            "-m",
            "neuron_morphology.transforms.pia_wm_streamlines."
            "calculate_pia_wm_streamlines",
        ]
        for key, value in self.input.items():
            cmd.append(f"--{key}")
            cmd.append(f"{value}")

        sp.check_call(cmd)

        depth_file = os.path.join(self.test_dir, "depth_field.nc")
        with xr.open_dataarray(depth_file) as da:
            self.assertEqual(len(da.x), 122)
            self.assertAlmostEqual(round(float(da[50, 50]), 5), 0.52746)
            self.assertEqual(np.isnan(da[0, 0]), True)

        gradient_file = os.path.join(self.test_dir, "gradient_field.nc")
        with xr.open_dataarray(gradient_file) as dg:
            self.assertEqual(len(dg.y), 102)
            x = dg[50, 70].values
            x = x / np.linalg.norm(x)
            assert np.allclose(x[1], 1, atol=1e-2)
            self.assertEqual(np.all(np.isnan(dg[0, 0])), True)



import unittest
from neuron_morphology.validation import validate_reconstruction


class TestValidateReconstructionFunctions(unittest.TestCase):
    """ Tests the functions in validate_reconstruction.py """

    def test_parser_with_one_swc(self):
        parser = validate_reconstruction.parse_arguments(['--swc', 'tests.swc'])
        self.assertTrue(parser.swc, ["tests.swc"])

    def test_parser_with_one_marker(self):
        parser = validate_reconstruction.parse_arguments(['--marker', 'tests.marker'])
        self.assertTrue(parser.marker, ["tests.marker"])

    def test_parser_without_argument(self):
        parser = validate_reconstruction.parse_arguments([])
        self.assertEqual(parser.dir, None)
        self.assertEqual(parser.swc, None)
        self.assertEqual(parser.swc, None)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestValidateReconstructionFunctions)
    unittest.TextTestRunner(verbosity=5).run(suite)


import sys
import unittest

import numpy as np


try:
    import dolfinx
    from neuron_morphology.transforms import streamline as sl

    has_dolfinx = True
except ImportError:
    has_dolfinx = False


@unittest.skipIf(
    not has_dolfinx, "streamline calculation requires conda-installed fenics-dolfinx"
)
class TestLaplace(unittest.TestCase):
    def test_generate_laplace_field_mesh_values(self):
        pia = [(1, 1), (0, 1)]
        wm = [(0, 0), (1, 0)]

        (
            u,
            grad_u,
            mesh,
            mesh_coords,
            mesh_values,
            mesh_gradients,
        ) = sl.generate_laplace_field(pia, wm, mesh_res=1)
        print(mesh_coords)
        assert np.allclose(mesh_values, mesh_coords[:,1])

    def test_generate_laplace_field_mesh_gradients(self):
        pia = [(1, 1), (0, 1)]
        wm = [(0, 0), (1, 0)]

        (
            u,
            grad_u,
            mesh,
            mesh_coords,
            mesh_values,
            mesh_gradients,
        ) = sl.generate_laplace_field(pia, wm, mesh_res=1)
        print(mesh_coords)
        assert np.allclose(mesh_gradients, [[0, 1] for i in range(len(mesh_gradients))])



import unittest
from neuron_morphology.constants import *
from neuron_morphology.morphology import Morphology
from tests.objects import (test_node,
                           test_morphology_small,
                           test_morphology_small_branching,
                           test_morphology_small_multiple_trees,
                           test_morphology_large,
                           )


class TestTree(unittest.TestCase):

    def test_get_root(self):

        morphology = test_morphology_small()
        expected_root = test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1)
        self.assertEqual(expected_root, morphology.get_root())

    def test_len(self):

        morphology = test_morphology_small()
        expected_length = 7
        self.assertEqual(expected_length, len(morphology))

    def test_children_of(self):

        morphology = test_morphology_small()
        node = test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1)
        children = morphology.children_of(node)
        expected_children = [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                             test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                             test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1)]
        self.assertEqual(children, expected_children)

    def test_parent_of(self):

        morphology = test_morphology_small()
        node = test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)
        expected_parent = test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1)
        parent = morphology.parent_of(node)
        self.assertEqual(expected_parent, parent)

    def test_get_children_of_node_by_type(self):

        morphology = test_morphology_small()
        node = test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1)
        expected_children_by_type = {BASAL_DENDRITE: [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)],
                                     APICAL_DENDRITE: [test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1)],
                                     AXON: [test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1)]}
        for node_type in [BASAL_DENDRITE, APICAL_DENDRITE, AXON]:
            child = morphology.get_children_of_node_by_types(node, [node_type])
            self.assertEqual(expected_children_by_type[node_type], child)

    def test_node_by_id(self):

        morphology = test_morphology_small()
        expected_node = test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1)
        node_by_id = morphology.node_by_id(4)
        self.assertEqual(expected_node, node_by_id)

    def test_get_root_id(self):

        morphology = test_morphology_small()
        expected_root_id = 1
        root_id = morphology.get_root_id()
        self.assertEqual(expected_root_id, root_id)

    def test_get_number_of_trees(self):

        morphology = test_morphology_small_multiple_trees()
        expected_number_of_trees = 2
        number_of_trees = morphology.get_number_of_trees()
        self.assertEqual(expected_number_of_trees, number_of_trees)

    def test_get_tree_list(self):

        morphology = test_morphology_small_multiple_trees()
        expected_trees = [[test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                           test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                           test_node(id=3, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                           test_node(id=4, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1)
                           ],
                          [test_node(id=5, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=-1),
                           test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=5),
                           test_node(id=7, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=6),
                           test_node(id=8, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=7),
                           test_node(id=9, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=8)
                           ]]
        trees = morphology.get_tree_list()
        self.assertEqual(expected_trees, trees)

    def test_get_tree_root(self):

        morphology = test_morphology_small_multiple_trees()
        root_by_number = {0: test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                          1: test_node(id=5, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=-1)}
        for tree_number in [0, 1]:
            root = morphology.get_root_for_tree(tree_number)
            self.assertEqual(root_by_number[tree_number], root)

    def test_get_node_by_types_basal(self):

        morphology = test_morphology_large()
        expected_nodes = [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                          test_node(id=3, type=BASAL_DENDRITE, x=430, y=630, z=20, radius=3, parent_node_id=2),
                          test_node(id=4, type=BASAL_DENDRITE, x=460, y=660, z=30, radius=3, parent_node_id=3),
                          test_node(id=5, type=BASAL_DENDRITE, x=490, y=690, z=40, radius=3, parent_node_id=4)]

        nodes = morphology.get_node_by_types([BASAL_DENDRITE])
        self.assertEqual(expected_nodes, nodes)

    def test_get_node_by_types_apical(self):

        morphology = test_morphology_large()
        expected_nodes = [test_node(id=6, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                          test_node(id=7, type=APICAL_DENDRITE, x=630, y=330, z=30, radius=3, parent_node_id=6),
                          test_node(id=8, type=APICAL_DENDRITE, x=660, y=360, z=40, radius=3, parent_node_id=7),
                          test_node(id=9, type=APICAL_DENDRITE, x=690, y=390, z=50, radius=3, parent_node_id=8),
                          test_node(id=10, type=APICAL_DENDRITE, x=710, y=420, z=60, radius=3, parent_node_id=9),
                          test_node(id=11, type=APICAL_DENDRITE, x=740, y=450, z=70, radius=3, parent_node_id=10)]

        nodes = morphology.get_node_by_types([APICAL_DENDRITE])
        self.assertEqual(expected_nodes, nodes)

    def test_get_node_by_types_axon(self):

        morphology = test_morphology_large()
        expected_nodes = [test_node(id=12, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1),
                          test_node(id=13, type=AXON, x=930, y=630, z=40, radius=3, parent_node_id=12),
                          test_node(id=14, type=AXON, x=960, y=660, z=50, radius=3, parent_node_id=13),
                          test_node(id=15, type=AXON, x=990, y=690, z=60, radius=3, parent_node_id=14),
                          test_node(id=16, type=AXON, x=1020, y=720, z=70, radius=3, parent_node_id=15),
                          test_node(id=17, type=AXON, x=1050, y=750, z=80, radius=3, parent_node_id=16)]

        nodes = morphology.get_node_by_types([AXON])
        self.assertEqual(expected_nodes, nodes)

    def test_get_non_soma_nodes(self):

        morphology = test_morphology_small()
        expected_nodes = [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                          test_node(id=3, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=2),
                          test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                          test_node(id=5, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=4),
                          test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1),
                          test_node(id=7, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=6)]

        nodes = morphology.get_non_soma_nodes()
        self.assertEqual(expected_nodes, nodes)

    def test_get_max_id(self):

        morphology = test_morphology_large()
        expected_id = 17
        max_id = morphology.get_max_id()
        self.assertEqual(expected_id, max_id)

    def test_get_segment_list(self):

        morphology = test_morphology_large()
        segments = morphology.get_segment_list()
        expected_segments = [[test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                              test_node(id=3, type=BASAL_DENDRITE, x=430, y=630, z=20, radius=3, parent_node_id=2),
                              test_node(id=4, type=BASAL_DENDRITE, x=460, y=660, z=30, radius=3, parent_node_id=3),
                              test_node(id=5, type=BASAL_DENDRITE, x=490, y=690, z=40, radius=3, parent_node_id=4)],
                             [test_node(id=6, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                              test_node(id=7, type=APICAL_DENDRITE, x=630, y=330, z=30, radius=3, parent_node_id=6),
                              test_node(id=8, type=APICAL_DENDRITE, x=660, y=360, z=40, radius=3, parent_node_id=7),
                              test_node(id=9, type=APICAL_DENDRITE, x=690, y=390, z=50, radius=3, parent_node_id=8),
                              test_node(id=10, type=APICAL_DENDRITE, x=710, y=420, z=60, radius=3, parent_node_id=9),
                              test_node(id=11, type=APICAL_DENDRITE, x=740, y=450, z=70, radius=3, parent_node_id=10)],
                             [test_node(id=12, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1),
                              test_node(id=13, type=AXON, x=930, y=630, z=40, radius=3, parent_node_id=12),
                              test_node(id=14, type=AXON, x=960, y=660, z=50, radius=3, parent_node_id=13),
                              test_node(id=15, type=AXON, x=990, y=690, z=60, radius=3, parent_node_id=14),
                              test_node(id=16, type=AXON, x=1020, y=720, z=70, radius=3, parent_node_id=15),
                              test_node(id=17, type=AXON, x=1050, y=750, z=80, radius=3, parent_node_id=16)]]
        self.assertEqual(expected_segments, segments)

    def test_get_compartment_list(self):

        morphology = test_morphology_small()
        compartments = morphology.compartments
        expected_compartments = [[test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                                  test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)],
                                 [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1),
                                  test_node(id=3, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=2)],
                                 [test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                                  test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1)],
                                 [test_node(id=4, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3, parent_node_id=1),
                                  test_node(id=5, type=APICAL_DENDRITE, x=600, y=300, z=20, radius=3,
                                            parent_node_id=4)],
                                 [test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                                  test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1)],
                                 [test_node(id=6, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=1),
                                  test_node(id=7, type=AXON, x=900, y=600, z=30, radius=3, parent_node_id=6)]]

        self.assertEqual(expected_compartments, compartments)

    def test_get_compartment_for_node(self):

        morphology = test_morphology_small()
        node = test_node(id=3, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=2)
        compartment = morphology.get_compartment_for_node(node)
        expected_compartment = [test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1), node]
        self.assertEqual(expected_compartment, compartment)

    def test_get_compartment_for_node_with_type(self):

        morphology = test_morphology_small()
        node_types = [AXON]
        node = test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)
        compartment = morphology.get_compartment_for_node(node, node_types)
        expected_compartment = None
        self.assertEqual(expected_compartment, compartment)

    def test_get_compartment_for_node_with_no_type(self):

        morphology = test_morphology_small()
        node = test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)
        compartment = morphology.get_compartment_for_node(node)
        expected_compartment = [test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1), node]
        self.assertEqual(expected_compartment, compartment)

    def test_get_compartment_length(self):

        morphology = test_morphology_small()
        compartment = [test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                       test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)]
        expected_legnth = 400.6245124802026
        length = morphology.get_compartment_length(compartment)
        self.assertEqual(expected_legnth, length)


    def test_get_compartment_surface_area(self):
        morphology = test_morphology_small()  # not actually using nodes

        compartment = [
            {"x": 0, "y": 0, "z": 0, "radius": 10},
            {"x": 0, "y": 0, "z": 5, "radius": 5}
        ]

        self.assertAlmostEqual(
            morphology.get_compartment_surface_area(compartment),
            333.21622,
            places=6
        )

    def test_get_compartment_volume(self):
        morphology = test_morphology_small()  # not actually using nodes

        compartment = [
            {"x": 0, "y": 0, "z": 0, "radius": 10},
            {"x": 0, "y": 0, "z": 5, "radius": 5}
        ]

        self.assertAlmostEqual(
            morphology.get_compartment_volume(compartment),
            916.297857,
            places=6
        )

    def test_get_compartment_midpoint(self):

        morphology = test_morphology_small()
        compartment = [test_node(id=1, type=SOMA, x=800, y=610, z=30, radius=35, parent_node_id=-1),
                       test_node(id=2, type=BASAL_DENDRITE, x=400, y=600, z=10, radius=3, parent_node_id=1)]
        midpoint = morphology.get_compartment_midpoint(compartment)
        expected_midpoint = [600.0, 605.0, 20.0]
        self.assertEqual(expected_midpoint, midpoint)

    def test_get_leaf_nodes(self):

        morphology = test_morphology_small()
        leaf_nodes = morphology.get_leaf_nodes()
        expected_leaf_node_ids = set([3, 5, 7])
        self.assertEqual(set([leaf_node['id'] for leaf_node in leaf_nodes]),
                         expected_leaf_node_ids)

    def test_get_branching_nodes(self):

        morphology = test_morphology_small_branching()
        branching_nodes = morphology.get_branching_nodes()
        expected_branching_node_ids = set([3, 7, 11])
        self.assertEqual(set([branching_node['id']
                              for branching_node in branching_nodes]),
                         expected_branching_node_ids)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTree)
    unittest.TextTestRunner(verbosity=5).run(suite)



import unittest

from neuron_morphology.feature_extractor.data import Data
from neuron_morphology.feature_extractor.mark import Mark
from neuron_morphology.feature_extractor.marked_feature import marked
from neuron_morphology.feature_extractor.feature_extraction_run import \
    FeatureExtractionRun
from neuron_morphology.morphology_builder import MorphologyBuilder

class TestFeatureExtractionRun(unittest.TestCase):
    
    def setUp(self):

        # we are not actually going to use this - but it must be argued to the 
        # Data constructor
        self.morphology = (
            MorphologyBuilder()
                .root()
                .build()
        )

        class AMark(Mark):
            @classmethod
            def validate(cls, data):
                return hasattr(data, "a") and data.a == 2
        self.amark = AMark

        class BMark(Mark):
            @classmethod
            def validate(cls, data):
                return hasattr(data, "b") and data.b == 3
        self.bmark = BMark

        @marked(AMark)
        def foo(data):
            return data.a == 2
        self.foo = foo

        @marked(BMark)
        def baz(data):
            return data.b == 3
        self.baz = baz


    def test_select_marks(self):
        run = (
            FeatureExtractionRun(Data(self.morphology, a=2, b=4))
                .select_marks([self.amark, self.bmark])
        )

        self.assertEqual(run.selected_marks, {self.amark})


    def test_select_features(self):
        run = (
            FeatureExtractionRun(Data(self.morphology, a=2, b=4))
                .select_marks([self.amark, self.bmark])
                .select_features([self.foo, self.baz])
        )
        self.assertEqual(set(run.selected_features), {self.foo})

    def test_select_features_only(self):
        run = (
            FeatureExtractionRun(Data(self.morphology, a=2, b=4))
                .select_marks([self.amark, self.bmark])
                .select_features([self.foo, self.baz], only_marks={self.bmark})
        )
        self.assertEqual(set(run.selected_features), set())

    def test_extract(self):
        run = FeatureExtractionRun(Data(self.morphology, a=2, b=4))
        run.selected_features = [self.foo]
        run.extract()

        self.assertEqual(run.results["foo"], True)
        self.assertEqual(len(run.results), 1)
